import logging
from typing import Optional

from fastapi import Depends, HTTPException, Response, status
from fastapi.security import APIKeyHeader, HTTPAuthorizationCredentials, HTTPBearer
from firebase_admin import auth as firebase_auth
from pydantic import ValidationError
from sqlmodel import Session

from studio.app.common.core.auth.auth_config import AUTH_CONFIG
from studio.app.common.core.auth.security import validate_access_token
from studio.app.common.db.database import get_db
from studio.app.common.models import User as UserModel
from studio.app.common.models import UserRole as UserRoleModel
from studio.app.common.schemas.users import User


async def get_current_user(
    res: Response,
    ex_token: Optional[str] = Depends(APIKeyHeader(name="ExToken", auto_error=False)),
    credential: HTTPAuthorizationCredentials = Depends(HTTPBearer(auto_error=False)),
    db: Session = Depends(get_db),
):
    use_firebase_auth = AUTH_CONFIG.USE_FIREBASE_TOKEN
    try:
        assert credential is not None if use_firebase_auth else True
        assert ex_token is not None if not use_firebase_auth else True

        uid = None
        if use_firebase_auth:
            user = firebase_auth.verify_id_token(credential.credentials)
            uid = user["uid"]
        else:
            payload, err = validate_access_token(ex_token)
            assert err is None, str(err)
            uid = payload["sub"]

        user_data = (
            db.query(UserModel, UserRoleModel.role_id)
            .outerjoin(UserRoleModel, UserRoleModel.user_id == UserModel.id)
            .filter(UserModel.uid == uid)
            .first()
        )
        assert user_data is not None, "Invalid user data"
        authed_user, role_id = user_data
        authed_user.__dict__["role_id"] = role_id
        return User.from_orm(authed_user)

    except ValidationError as e:
        logging.getLogger().error(e)
        raise HTTPException(status_code=422, detail=f"Validator Error: {e}")
    except Exception as e:
        logging.getLogger().error(e)
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            headers={"WWW-Authenticate": 'Bearer realm="auth_required"'},
            detail=str(e) or "Could not validate credentials",
        )


async def get_admin_user(current_user: User = Depends(get_current_user)):
    if current_user.is_admin:
        return current_user
    else:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Insufficient privileges",
        )



from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Tuple, Union

from jose import ExpiredSignatureError, JWTError, jwt
from pydantic import ValidationError

from studio.app.common.core.auth.auth_config import AUTH_CONFIG


def _create_token(
    subject: Union[str, Any],
    token_type: str,
    expires_delta: timedelta = None,
    user_claims: Dict[str, Any] = None,
) -> str:
    token_data = {"sub": subject, "token_type": token_type}

    token_data.update(
        {"exp": datetime.utcnow() + expires_delta} if expires_delta else {}
    )
    token_data.update(user_claims if user_claims else {})

    return jwt.encode(
        token_data,
        AUTH_CONFIG.SECRET_KEY,
        algorithm=AUTH_CONFIG.ALGORITHM,
    )


def create_access_token(
    subject: Union[str, Any],
    expires_delta: timedelta = None,
    user_claims: Dict[str, Any] = None,
) -> str:
    return _create_token(
        subject,
        token_type="access_token",
        expires_delta=expires_delta,
        user_claims=user_claims,
    )


def create_refresh_token(
    subject: Union[str, Any],
    expires_delta: timedelta = None,
    user_claims: Dict[str, Any] = None,
) -> str:
    exp = (
        expires_delta
        if expires_delta
        else timedelta(minutes=AUTH_CONFIG.REFRESH_TOKEN_EXPIRE_MINUTES)
    )
    return _create_token(
        subject,
        token_type="refresh_token",
        expires_delta=exp,
        user_claims=user_claims,
    )


def _validate_token(
    token: str,
    token_type: str,
) -> Tuple[Dict[str, Any], Optional[str]]:
    payload = err = None
    try:
        payload = jwt.decode(
            token, AUTH_CONFIG.SECRET_KEY, algorithms=[AUTH_CONFIG.ALGORITHM]
        )
        if payload.get("token_type") != token_type:
            err = "Invalid token type"
    except ExpiredSignatureError:
        err = "Credentials has expired"
    except (JWTError, ValidationError):
        err = "Could not validate credentials"
    except Exception:
        err = "Bad token"

    return payload, err


def validate_access_token(token: str) -> Tuple[Dict[str, Any], Optional[str]]:
    return _validate_token(token, "access_token")


def validate_refresh_token(token: str) -> Tuple[Dict[str, Any], Optional[str]]:
    return _validate_token(token, "refresh_token")



from studio.app.optinist.wrappers.caiman import caiman_wrapper_dict
from studio.app.optinist.wrappers.lccd import lccd_wrapper_dict
from studio.app.optinist.wrappers.optinist import optinist_wrapper_dict
from studio.app.optinist.wrappers.suite2p import suite2p_wrapper_dict

wrapper_dict = {}
wrapper_dict.update(**caiman_wrapper_dict)
wrapper_dict.update(**suite2p_wrapper_dict)
wrapper_dict.update(**lccd_wrapper_dict)
wrapper_dict.update(**optinist_wrapper_dict)



from typing import Optional

from pydantic import BaseModel, EmailStr


class Token(BaseModel):
    access_token: str
    token_type: str
    refresh_token: Optional[str]
    ex_token: Optional[str]


class RefreshToken(BaseModel):
    refresh_token: str


class UserAuth(BaseModel):
    email: EmailStr
    password: str


class AccessToken(BaseModel):
    access_token: str



from datetime import datetime
from typing import List, Optional

from sqlalchemy.dialects.mysql import BIGINT
from sqlalchemy.sql.functions import current_timestamp
from sqlmodel import Column, Field, ForeignKey, Relationship, String, UniqueConstraint

from studio.app.common.models.base import Base, TimestampMixin


class WorkspacesShareUser(Base, table=True):
    __tablename__ = "workspaces_share_users"
    __table_args__ = (
        UniqueConstraint("workspace_id", "user_id", name="idx_workspace_id_user_id"),
    )

    workspace_id: int = Field(
        sa_column=Column(
            BIGINT(unsigned=True), ForeignKey("workspaces.id"), nullable=False
        ),
    )
    user_id: int = Field(
        sa_column=Column(BIGINT(unsigned=True), ForeignKey("users.id"), nullable=False),
    )
    created_at: Optional[datetime] = Field(
        sa_column_kwargs={"server_default": current_timestamp()},
    )


class Workspace(Base, TimestampMixin, table=True):
    __tablename__ = "workspaces"

    name: str = Field(sa_column=Column(String(100), nullable=False))
    user_id: int = Field(
        sa_column=Column(
            BIGINT(unsigned=True), ForeignKey("users.id", name="user"), nullable=False
        ),
    )
    deleted: bool = Field(nullable=False)

    user: Optional["User"] = Relationship(back_populates="workspace")  # noqa: F821
    user_share: List["User"] = Relationship(  # noqa: F821
        back_populates="workspace_share", link_model=WorkspacesShareUser
    )



from studio.app.optinist.core.edit_ROI.wrappers.lccd_edit_roi.add_roi import (
    execute_add_ROI,
)
from studio.app.optinist.core.edit_ROI.wrappers.lccd_edit_roi.delete_roi import (
    excute_delete_roi,
)
from studio.app.optinist.core.edit_ROI.wrappers.lccd_edit_roi.merge_roi import (
    execute_merge_roi,
)

__all__ = ["execute_add_ROI", "excute_delete_roi", "execute_merge_roi"]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
from datetime import datetime

from sphinx_pyproject import SphinxConfig

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#


# -- Project information -----------------------------------------------------
config = SphinxConfig("../pyproject.toml", globalns=globals())
project = "OptiNiSt"
copyright = f"{datetime.today().year}, OIST"
author = ""
release = ".".join(config.version.split(".")[:2])

# -- readthedocs -------------------------------------------------------------
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# -- General configuration ---------------------------------------------------
master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.coverage",
    "sphinxcontrib.apidoc",
    "sphinx_autodoc_typehints",
    "myst_parser",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.autodoc.typehints",
]

# Tell myst-parser to assign header anchors for h1-h3.
myst_heading_anchors = 4

suppress_warnings = ["myst.header"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "tests", "*test*"]

source_suffix = [".rst", ".md"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_logo = "_static/optinist.png"
html_favicon = "_static/favicon.ico"

# disable document page source link
html_show_sourcelink = False

autosummary_generate = True

html_theme_options = {
    "canonical_url": "",
    "logo_only": False,
    "display_version": True,
    "prev_next_buttons_location": "top",
    "style_external_links": False,
    "style_nav_header_background": "#C3EBE1",
    # Toc options
    "collapse_navigation": True,
    "sticky_navigation": True,
    "navigation_depth": 4,
    "includehidden": True,
    "titles_only": False,
}



from studio.app.common.core.experiment.experiment import ExptConfig


class ExptConfigBuilder:
    def __init__(self):
        self._workspace_id = None
        self._unique_id = None
        self._name = None
        self._started_at = None
        self._finished_at = None
        self._success = None
        self._hasNWB = False
        self._function = {}
        self._nwbfile = None
        self._snakemake = None

    def set_config(self, config: ExptConfig) -> "ExptConfigBuilder":
        self._workspace_id = config.workspace_id
        self._unique_id = config.unique_id
        self._name = config.name
        self._started_at = config.started_at
        self._finished_at = config.finished_at
        self._success = config.success
        self._hasNWB = config.hasNWB
        self._function = config.function
        self._nwbfile = config.nwb
        self._snakemake = config.snakemake
        return self

    def set_workspace_id(self, workspace_id) -> "ExptConfigBuilder":
        self._workspace_id = workspace_id
        return self

    def set_unique_id(self, unique_id) -> "ExptConfigBuilder":
        self._unique_id = unique_id
        return self

    def set_name(self, name) -> "ExptConfigBuilder":
        self._name = name
        return self

    def set_started_at(self, timestamp) -> "ExptConfigBuilder":
        self._started_at = timestamp
        return self

    def set_success(self, success: str) -> "ExptConfigBuilder":
        self._success = success
        return self

    def set_hasNWB(self, hasNWB) -> "ExptConfigBuilder":
        self._hasNWB = hasNWB
        return self

    def set_function(self, function) -> "ExptConfigBuilder":
        self._function = function
        return self

    def set_nwbfile(self, nwbfile) -> "ExptConfigBuilder":
        self._nwbfile = nwbfile
        return self

    def set_snakemake(self, snakemake) -> "ExptConfigBuilder":
        self._snakemake = snakemake
        return self

    def build(self) -> ExptConfig:
        return ExptConfig(
            workspace_id=self._workspace_id,
            unique_id=self._unique_id,
            name=self._name,
            started_at=self._started_at,
            finished_at=self._finished_at,
            success=self._success,
            hasNWB=self._hasNWB,
            function=self._function,
            nwb=self._nwbfile,
            snakemake=self._snakemake,
        )



from typing import Union

from fastapi import Depends, HTTPException
from sqlmodel import Session, or_

from studio.app.common import models as common_model
from studio.app.common.core.auth.auth_dependencies import get_current_user
from studio.app.common.db.database import get_db
from studio.app.common.schemas.users import User


async def is_workspace_owner(
    workspace_id: Union[int, str],
    current_user: User = Depends(get_current_user),
    db: Session = Depends(get_db),
):
    workspace = (
        db.query(common_model.Workspace)
        .filter(
            common_model.Workspace.id == workspace_id,
            common_model.Workspace.deleted.is_(False),
            common_model.Workspace.user_id == current_user.id,
        )
        .first()
    )

    if workspace is None:
        raise HTTPException(status_code=403, detail="Operation is not available")
    else:
        return True


async def is_workspace_available(
    workspace_id: Union[int, str],
    current_user: User = Depends(get_current_user),
    db: Session = Depends(get_db),
) -> bool:
    workspace = (
        db.query(common_model.Workspace)
        .join(
            common_model.WorkspacesShareUser,
            common_model.WorkspacesShareUser.workspace_id == common_model.Workspace.id,
            isouter=True,
        )
        .filter(
            common_model.Workspace.id == workspace_id,
            common_model.Workspace.deleted.is_(False),
            or_(
                common_model.Workspace.user_id == current_user.id,
                common_model.WorkspacesShareUser.user_id == current_user.id,
            ),
        )
        .group_by(common_model.Workspace.id)
        .first()
    )

    if workspace is None:
        raise HTTPException(status_code=403, detail="Workspace is not available")
    else:
        return True



import copy
import gc
import os
import traceback
from dataclasses import asdict
from datetime import datetime

from filelock import FileLock

from studio.app.common.core.experiment.experiment_reader import ExptConfigReader
from studio.app.common.core.snakemake.smk import Rule
from studio.app.common.core.utils.config_handler import ConfigWriter
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.utils.pickle_handler import PickleReader, PickleWriter
from studio.app.const import DATE_FORMAT
from studio.app.dir_path import DIRPATH
from studio.app.optinist.core.nwb.nwb_creater import (
    merge_nwbfile,
    overwrite_nwbfile,
    save_nwb,
)
from studio.app.wrappers import wrapper_dict


class Runner:
    @classmethod
    def run(cls, __rule: Rule, last_output):
        try:
            input_info = cls.read_input_info(__rule.input)

            cls.change_dict_key_exist(input_info, __rule)

            nwbfile = input_info["nwbfile"]

            # input_info
            for key in list(input_info):
                if key not in __rule.return_arg.values():
                    input_info.pop(key)

            cls.set_func_start_timestamp(os.path.dirname(__rule.output))

            # output_info
            output_info = cls.execute_function(
                __rule.path,
                __rule.params,
                nwbfile.get("input"),
                os.path.dirname(__rule.output),
                input_info,
            )

            # nwbfileの設定
            output_info["nwbfile"] = cls.save_func_nwb(
                f"{__rule.output.split('.')[0]}.nwb",
                __rule.type,
                nwbfile,
                output_info,
            )

            # 各関数での結果を保存
            PickleWriter.write(__rule.output, output_info)

            # NWB全体保存
            if __rule.output in last_output:
                # 全体の結果を保存する
                path = join_filepath(os.path.dirname(os.path.dirname(__rule.output)))
                path = join_filepath([path, "whole.nwb"])
                cls.save_all_nwb(path, output_info["nwbfile"])

            print("output: ", __rule.output)

            del input_info, output_info
            gc.collect()

        except Exception as e:
            PickleWriter.write(
                __rule.output,
                list(traceback.TracebackException.from_exception(e).format())[-2:],
            )

    @classmethod
    def set_func_start_timestamp(cls, output_dirpath):
        workflow_dirpath = os.path.dirname(output_dirpath)
        node_id = os.path.basename(output_dirpath)
        expt_config = ExptConfigReader.read(
            join_filepath([workflow_dirpath, DIRPATH.EXPERIMENT_YML])
        )
        expt_config.function[node_id].started_at = datetime.now().strftime(DATE_FORMAT)
        ConfigWriter.write(
            dirname=workflow_dirpath,
            filename=DIRPATH.EXPERIMENT_YML,
            config=asdict(expt_config),
        )

    @classmethod
    def save_func_nwb(cls, save_path, name, nwbfile, output_info):
        if "nwbfile" in output_info:
            nwbfile[name] = output_info["nwbfile"]
            save_nwb(
                save_path,
                nwbfile["input"],
                output_info["nwbfile"],
            )
        return nwbfile

    @classmethod
    def save_all_nwb(cls, save_path, all_nwbfile):
        input_nwbfile = all_nwbfile["input"]
        all_nwbfile.pop("input")
        nwbconfig = {}
        for x in all_nwbfile.values():
            nwbconfig = merge_nwbfile(nwbconfig, x)
        # 同一のnwbfileに対して、複数の関数を実行した場合、h5pyエラーが発生する
        lock_path = save_path + ".lock"
        timeout = 30  # ロック取得のタイムアウト時間（秒）
        with FileLock(lock_path, timeout=timeout):
            # ロックが取得できたら、ファイルに書き込みを行う
            try:
                if os.path.exists(save_path):
                    overwrite_nwbfile(save_path, nwbconfig)
                else:
                    save_nwb(save_path, input_nwbfile, nwbconfig)
            except Exception as e:
                print(e)

    @classmethod
    def execute_function(cls, path, params, nwb_params, output_dir, input_info):
        wrapper = cls.dict2leaf(wrapper_dict, path.split("/"))
        func = copy.deepcopy(wrapper["function"])
        output_info = func(
            params=params, nwbfile=nwb_params, output_dir=output_dir, **input_info
        )
        del func
        gc.collect()

        return output_info

    @classmethod
    def change_dict_key_exist(cls, input_info, rule_config: Rule):
        for return_name, arg_name in rule_config.return_arg.items():
            if return_name in input_info:
                input_info[arg_name] = input_info.pop(return_name)

    @classmethod
    def read_input_info(cls, input_files):
        input_info = {}
        for filepath in input_files:
            load_data = PickleReader.read(filepath)
            input_info = dict(list(load_data.items()) + list(input_info.items()))
        return input_info

    @classmethod
    def dict2leaf(cls, root_dict: dict, path_list):
        path = path_list.pop(0)
        if len(path_list) > 0:
            return cls.dict2leaf(root_dict[path], path_list)
        else:
            return root_dict[path]



from pydantic import BaseSettings, Field

from studio.app.dir_path import DIRPATH


class AuthConfig(BaseSettings):
    REFRESH_TOKEN_EXPIRE_MINUTES: int = Field(
        default=60 * 24 * 1, env="REFRESH_TOKEN_EXPIRE_MINUTES"
    )
    SECRET_KEY: str = Field(default="123456", env="SECRET_KEY")
    USE_FIREBASE_TOKEN: bool = Field(default=True, env="USE_FIREBASE_TOKEN")
    ALGORITHM = "HS256"

    class Config:
        env_file = f"{DIRPATH.CONFIG_DIR}/.env"
        env_file_encoding = "utf-8"


AUTH_CONFIG = AuthConfig()



import logging
import os
from typing import Dict

from studio.app.common.core.utils.file_reader import JsonReader
from studio.app.common.core.utils.filepath_creater import (
    create_directory,
    join_filepath,
)
from studio.app.dir_path import DIRPATH


def get_logger(workspace_id: str, unique_id: str) -> logging.Logger:
    output_dirpath = join_filepath(
        [
            DIRPATH.OUTPUT_DIR,
            workspace_id,
            unique_id,
        ]
    )
    create_directory(output_dirpath)

    filepath = f"{output_dirpath}/error.log"
    if os.path.exists(filepath):
        try:
            os.remove(filepath)
        except Exception as e:
            print("[Exception][Logger]", e)

    logger = logging.getLogger(unique_id)

    # FileHandlerの設定
    fh = logging.FileHandler(filepath)
    fh.setLevel(logging.ERROR)
    fmt = logging.Formatter("%(asctime)s : %(levelname)s - %(filename)s - %(message)s")
    fh.setFormatter(fmt)
    logger.addHandler(fh)
    fh.close()
    return logger


class Logger:
    # TODO: Since the Snakemake library automatically creates thread for workflows and
    # shares the same loggers in the library, all workflows running at the same time
    # will use same log data.
    def __init__(self, workspace_id, unique_id):
        self.workspace_id = workspace_id
        self.unique_id = unique_id
        self.logger: logging.Logger = get_logger(workspace_id, unique_id)

    def smk_logger(self, msg: Dict[str, str] = None):
        """
        msg:
            level:
                "job_info": jobの始まりを通知。inputやoutputがある
                "job_finished": jobの終了を通知。
        """
        # pass
        # # エラーした
        # if "exception" in msg:
        #     self.logger.error(msg)

        if "level" in msg and "debug" in msg["level"]:
            level = msg["level"]
            if "debug" in level and "msg" in msg:
                if "Traceback" in msg["msg"]:
                    # check if the message is thrown by killing process action
                    if any(
                        err in msg["msg"]
                        for err in ["Signals.SIGTERM", "exit status 15"]
                    ):
                        pid_filepath = join_filepath(
                            [
                                DIRPATH.OUTPUT_DIR,
                                self.workspace_id,
                                self.unique_id,
                                "pid.json",
                            ]
                        )
                        pid_data = JsonReader.read(pid_filepath)
                        # since multiple running workflow share log data,
                        # check if message really belongs to the current workflow
                        if pid_data["last_script_file"] in msg["msg"]:
                            self.logger.error("Workflow cancelled")
                    else:
                        self.logger.error(msg)

    def clean_up(self):
        """
        remove all handlers from this logger
        """
        self.logger.handlers.clear()



import os
from enum import Enum

_DEFAULT_DIR = "/tmp/studio"
_ENV_DIR = os.environ.get("OPTINIST_DIR")


class DIRPATH:
    DATA_DIR = _DEFAULT_DIR if _ENV_DIR is None else _ENV_DIR

    INPUT_DIR = f"{DATA_DIR}/input"
    OUTPUT_DIR = f"{DATA_DIR}/output"

    if not os.path.exists(INPUT_DIR):
        os.makedirs(INPUT_DIR)
    assert os.path.exists(INPUT_DIR)

    if not os.path.exists(OUTPUT_DIR):
        os.makedirs(OUTPUT_DIR)
    assert os.path.exists(OUTPUT_DIR)

    ROOT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    STUDIO_DIR = os.path.dirname(os.path.dirname(__file__))
    APP_DIR = os.path.dirname(__file__)
    CONFIG_DIR = f"{STUDIO_DIR}/config"

    CONDAENV_DIR = (
        f"{os.path.dirname(os.path.dirname(os.path.dirname(__file__)))}/conda"
    )

    SNAKEMAKE_FILEPATH = f"{APP_DIR}/Snakefile"
    EXPERIMENT_YML = "experiment.yaml"
    SNAKEMAKE_CONFIG_YML = "snakemake.yaml"
    WORKFLOW_YML = "workflow.yaml"

    FIREBASE_PRIVATE_PATH = f"{CONFIG_DIR}/auth/firebase_private.json"
    FIREBASE_CONFIG_PATH = f"{CONFIG_DIR}/auth/firebase_config.json"


class CORE_PARAM_PATH(Enum):
    nwb = f"{DIRPATH.APP_DIR}/optinist/core/nwb/nwb.yaml"
    snakemake = f"{DIRPATH.APP_DIR}/common/core/snakemake/snakemake.yaml"



from datetime import datetime
from glob import glob
from typing import Optional

from studio.app.common.core.experiment.experiment import ExptConfig
from studio.app.common.core.experiment.experiment_reader import ExptConfigReader
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.const import DATE_FORMAT
from studio.app.dir_path import DIRPATH


class ExptUtils:
    @classmethod
    def get_last_experiment(cls, workspace_id: str):
        last_expt_config: Optional[ExptConfig] = None
        config_paths = glob(
            join_filepath(
                [DIRPATH.OUTPUT_DIR, workspace_id, "*", DIRPATH.EXPERIMENT_YML]
            )
        )

        for path in config_paths:
            config = ExptConfigReader.read(path)
            if not last_expt_config:
                last_expt_config = config
            elif datetime.strptime(config.started_at, DATE_FORMAT) > datetime.strptime(
                last_expt_config.started_at, DATE_FORMAT
            ):
                last_expt_config = config
        return last_expt_config



from datetime import datetime
from typing import Optional

from sqlalchemy.dialects.mysql import BIGINT
from sqlalchemy.sql.functions import current_timestamp
from sqlmodel import Column, Field, SQLModel, text


class Base(SQLModel):
    id: int = Field(sa_column=Column(BIGINT(unsigned=True), primary_key=True))


class TimestampMixin(SQLModel):
    created_at: Optional[datetime] = Field(
        sa_column_kwargs={"server_default": current_timestamp()},
    )

    updated_at: Optional[datetime] = Field(
        sa_column_kwargs={
            "server_default": text("CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP"),
        },
    )



from studio.app.optinist.dataclass.behavior import BehaviorData
from studio.app.optinist.dataclass.caiman import CaimanCnmfData
from studio.app.optinist.dataclass.fluo import FluoData
from studio.app.optinist.dataclass.iscell import IscellData
from studio.app.optinist.dataclass.lccd import LccdData
from studio.app.optinist.dataclass.nwb import NWBFile
from studio.app.optinist.dataclass.roi import RoiData
from studio.app.optinist.dataclass.suite2p import Suite2pData

__all__ = [
    "BehaviorData",
    "CaimanCnmfData",
    "FluoData",
    "IscellData",
    "LccdData",
    "NWBFile",
    "RoiData",
    "Suite2pData",
]



from fastapi import APIRouter, Depends, HTTPException
from fastapi_pagination import LimitOffsetPage
from fastapi_pagination.ext.sqlmodel import paginate
from sqlalchemy import func
from sqlmodel import Session, or_, select

from studio.app.common import models as common_model
from studio.app.common.core.auth.auth_dependencies import get_current_user
from studio.app.common.core.workspace.workspace_dependencies import (
    is_workspace_available,
    is_workspace_owner,
)
from studio.app.common.db.database import get_db
from studio.app.common.schemas.base import SortOptions
from studio.app.common.schemas.users import User
from studio.app.common.schemas.workspace import (
    Workspace,
    WorkspaceCreate,
    WorkspaceSharePostStatus,
    WorkspaceShareStatus,
    WorkspaceUpdate,
)

router = APIRouter(tags=["Workspace"])


shared_count_subquery = (
    select(func.count())
    .select_from(common_model.WorkspacesShareUser)
    .join(
        common_model.User,
        common_model.WorkspacesShareUser.user_id == common_model.User.id,
    )
    .where(
        common_model.WorkspacesShareUser.workspace_id == common_model.Workspace.id,
        common_model.User.active.is_(True),
    )
    .correlate(common_model.Workspace)
    .as_scalar()
)


@router.get(
    "/workspaces",
    response_model=LimitOffsetPage[Workspace],
    description="""
- search workspaces
""",
)
def search_workspaces(
    sortOptions: SortOptions = Depends(),
    db: Session = Depends(get_db),
    current_user: User = Depends(get_current_user),
):
    sa_sort_list = sortOptions.get_sa_sort_list(sa_table=common_model.Workspace)

    def workspace_transformer(items):
        list_ws = []
        for item in items:
            item[0].__dict__["shared_count"] = item.shared_count
            list_ws.append(item[0])
        return list_ws

    query = (
        select(common_model.Workspace, shared_count_subquery.label("shared_count"))
        .join(
            common_model.WorkspacesShareUser,
            common_model.Workspace.id == common_model.WorkspacesShareUser.workspace_id,
            isouter=True,
        )
        .filter(
            common_model.Workspace.deleted.is_(False),
            or_(
                common_model.WorkspacesShareUser.user_id == current_user.id,
                common_model.Workspace.user_id == current_user.id,
            ),
        )
        .group_by(common_model.Workspace.id)
        .order_by(*sa_sort_list)
    )

    data = paginate(
        db,
        query,
        transformer=workspace_transformer,
    )
    return data


@router.get(
    "/workspace/{workspace_id}",
    response_model=Workspace,
    dependencies=[Depends(is_workspace_available)],
    description="""
- get workspace by id
""",
)
def get_workspace(
    workspace_id: int,
    db: Session = Depends(get_db),
    current_user: User = Depends(get_current_user),
):
    data = (
        db.query(common_model.Workspace, shared_count_subquery.label("shared_count"))
        .outerjoin(
            common_model.WorkspacesShareUser,
            common_model.Workspace.id == common_model.WorkspacesShareUser.workspace_id,
        )
        .filter(
            common_model.Workspace.id == workspace_id,
            common_model.Workspace.deleted.is_(False),
            or_(
                common_model.WorkspacesShareUser.user_id == current_user.id,
                common_model.Workspace.user_id == current_user.id,
            ),
        )
        .first()
    )
    if not data:
        raise HTTPException(status_code=404)
    workspace, shared_count = data
    workspace.__dict__["shared_count"] = shared_count
    return Workspace.from_orm(workspace)


@router.post(
    "/workspace",
    response_model=Workspace,
    description="""
- create workspace
""",
)
def create_workspace(
    workspace: WorkspaceCreate,
    db: Session = Depends(get_db),
    current_user: User = Depends(get_current_user),
):
    workspace = common_model.Workspace(
        **workspace.dict(), user_id=current_user.id, deleted=0
    )
    db.add(workspace)
    db.commit()
    db.refresh(workspace)
    workspace.__dict__["user"] = current_user
    workspace.__dict__["shared_count"] = 0
    return Workspace.from_orm(workspace)


@router.put(
    "/workspace/{workspace_id}",
    response_model=Workspace,
    dependencies=[Depends(is_workspace_owner)],
    description="""
- update workspace
""",
)
def update_workspace(
    workspace_id: int,
    workspace: WorkspaceUpdate,
    db: Session = Depends(get_db),
    current_user: User = Depends(get_current_user),
):
    data = (
        db.query(common_model.Workspace, shared_count_subquery)
        .filter(
            common_model.Workspace.id == workspace_id,
            common_model.Workspace.user_id == current_user.id,
            common_model.Workspace.deleted.is_(False),
        )
        .first()
    )
    if not data:
        raise HTTPException(status_code=404)
    ws, shared_count = data
    data = workspace.dict(exclude_unset=True)
    for key, value in data.items():
        setattr(ws, key, value)
    db.commit()
    db.refresh(ws)
    ws.__dict__["user"] = current_user
    ws.__dict__["shared_count"] = shared_count
    return Workspace.from_orm(ws)


@router.delete(
    "/workspace/{workspace_id}",
    response_model=bool,
    dependencies=[Depends(is_workspace_owner)],
    description="""
- delete workspace
""",
)
def delete_workspace(
    workspace_id: int,
    db: Session = Depends(get_db),
    current_user: User = Depends(get_current_user),
):
    ws = (
        db.query(common_model.Workspace)
        .filter(
            common_model.Workspace.id == workspace_id,
            common_model.Workspace.user_id == current_user.id,
            common_model.Workspace.deleted.is_(False),
        )
        .first()
    )
    if not ws:
        raise HTTPException(status_code=404)
    ws.deleted = True
    db.commit()
    return True


@router.get(
    "/workspace/share/{workspace_id}/status",
    response_model=WorkspaceShareStatus,
    dependencies=[Depends(is_workspace_available)],
    description="""
- get workspace share status
""",
)
def get_workspace_share_status(
    workspace_id: int,
    db: Session = Depends(get_db),
    current_user: User = Depends(get_current_user),
):
    workspace = (
        db.query(common_model.Workspace)
        .filter(
            common_model.Workspace.id == workspace_id,
            common_model.Workspace.user_id == current_user.id,
            common_model.Workspace.deleted.is_(False),
        )
        .first()
    )
    if not workspace:
        raise HTTPException(status_code=404)
    users = (
        db.query(common_model.User)
        .join(
            common_model.WorkspacesShareUser,
            common_model.WorkspacesShareUser.user_id == common_model.User.id,
        )
        .filter(
            common_model.WorkspacesShareUser.workspace_id == workspace_id,
            common_model.User.active.is_(True),
        )
        .all()
    )
    return WorkspaceShareStatus(users=users)


@router.post(
    "/workspace/share/{workspace_id}/status",
    response_model=bool,
    dependencies=[Depends(is_workspace_owner)],
    description="""
- update workspace share status
""",
)
def update_workspace_share_status(
    workspace_id: int,
    data: WorkspaceSharePostStatus,
    db: Session = Depends(get_db),
    current_user: User = Depends(get_current_user),
):
    workspace = (
        db.query(common_model.Workspace)
        .filter(
            common_model.Workspace.id == workspace_id,
            common_model.Workspace.user_id == current_user.id,
            common_model.Workspace.deleted.is_(False),
        )
        .first()
    )
    if not workspace:
        raise HTTPException(status_code=404)

    (
        db.query(common_model.WorkspacesShareUser)
        .filter(common_model.WorkspacesShareUser.workspace_id == workspace_id)
        .delete(synchronize_session=False)
    )
    db.bulk_save_objects(
        common_model.WorkspacesShareUser(workspace_id=workspace_id, user_id=user_id)
        for user_id in data.user_ids
    )
    db.commit()
    return True



import numpy as np
import pandas as pd
import plotly.graph_objects as go
import plotly.io as pio

from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.utils.json_writer import JsonWriter
from studio.app.common.core.workflow.workflow import OutputPath, OutputType
from studio.app.common.dataclass.base import BaseData


class PieData(BaseData):
    def __init__(self, data, labels: list, file_name="pie"):
        super().__init__(file_name)

        if isinstance(data, list):
            data = np.array(data)
        assert isinstance(data, np.ndarray), "Pie Type Error"
        assert data.ndim == 1, "Pie Dimension Error"
        assert data.shape[0] == len(
            labels
        ), f"labels length is not same as data shape {data.shape}"
        self.data = data.reshape(1, -1)
        self.columns = labels

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.json"])
        df = pd.DataFrame(self.data, columns=self.columns)
        JsonWriter.write_as_split(self.json_path, df)

    @property
    def output_path(self) -> OutputPath:
        return OutputPath(path=self.json_path, type=OutputType.PIE)

    def save_plot(self, output_dir):
        fig = go.Figure(
            data=[
                go.Pie(
                    labels=self.columns,
                    values=self.data[0],
                    sort=False,
                    direction="clockwise",
                )
            ]
        )
        pio.write_image(fig, join_filepath([output_dir, f"{self.file_name}.png"]))



from fastapi import APIRouter, Depends
from sqlmodel import Session

from studio.app.common.core.auth import auth
from studio.app.common.db.database import get_db
from studio.app.common.schemas.auth import AccessToken, RefreshToken, Token, UserAuth

router = APIRouter(prefix="/auth", tags=["auth"])


@router.post("/login", response_model=Token)
async def login(user_data: UserAuth, db: Session = Depends(get_db)):
    return await auth.authenticate_user(db, user_data)


@router.post("/refresh", response_model=AccessToken)
async def refresh(refresh_token: RefreshToken):
    return await auth.refresh_current_user_token(refresh_token.refresh_token)


@router.post("/send_reset_password_mail")
async def send_reset_password_mail(email: str, db: Session = Depends(get_db)):
    return await auth.send_reset_password_mail(db, email)



from studio.app.common.dataclass.timeseries import TimeSeriesData


class FluoData(TimeSeriesData):
    def __init__(self, data, std=None, index=None, file_name="fluo"):
        super().__init__(
            data=data,
            std=std,
            index=index,
            file_name=file_name,
        )



# flake8: noqa
# Exclude from lint for the following reason
# This file is executed by snakemake and cause the following lint errors
# - E402: sys.path.append is required to import optinist modules
# - F821: do not import snakemake
import json
import os
import sys
from os.path import abspath, dirname
from pathlib import Path

ROOT_DIRPATH = dirname(dirname(dirname(dirname(dirname(dirname(abspath(__file__)))))))

sys.path.append(ROOT_DIRPATH)

from studio.app.common.core.rules.runner import Runner
from studio.app.common.core.snakemake.snakemake_reader import RuleConfigReader
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.dir_path import DIRPATH

if __name__ == "__main__":
    last_output = [
        join_filepath([DIRPATH.OUTPUT_DIR, x]) for x in snakemake.config["last_output"]
    ]

    rule_config = RuleConfigReader.read(snakemake.params.name)

    rule_config.input = snakemake.input
    rule_config.output = snakemake.output[0]

    # save snakemake script file path and PID of current running algo function
    pid_data = {"last_pid": os.getpid(), "last_script_file": sys.argv[0]}
    with open(Path(rule_config.output).parent.parent / "pid.json", "w") as f:
        json.dump(pid_data, f)

    Runner.run(rule_config, last_output)



from fastapi import HTTPException
from fastapi_pagination.ext.sqlmodel import paginate
from firebase_admin import auth as firebase_auth
from firebase_admin.auth import UserRecord
from sqlmodel import Session, select

from studio.app.common.core.auth.auth import authenticate_user
from studio.app.common.models import Role as RoleModel
from studio.app.common.models import User as UserModel
from studio.app.common.models import UserRole as UserRoleModel
from studio.app.common.schemas.auth import UserAuth
from studio.app.common.schemas.base import SortOptions
from studio.app.common.schemas.users import (
    User,
    UserCreate,
    UserPasswordUpdate,
    UserSearchOptions,
    UserUpdate,
)


async def set_role(db: Session, user_id: int, role_id: int, auto_commit=True):
    db.query(UserRoleModel).filter_by(user_id=user_id).delete(synchronize_session=False)
    role_user = UserRoleModel(user_id=user_id, role_id=role_id)
    db.add(role_user)
    db.flush()
    if auto_commit:
        db.commit()


async def get_user(db: Session, user_id: int, organization_id: int) -> User:
    try:
        data = (
            db.query(UserModel, UserRoleModel.role_id)
            .outerjoin(UserRoleModel, UserModel.id == UserRoleModel.user_id)
            .filter(
                UserModel.id == user_id,
                UserModel.active.is_(True),
                UserModel.organization_id == organization_id,
            )
            .first()
        )
        assert data is not None, "User not found"
        user, role_id = data
        user.__dict__["role_id"] = role_id
        return User.from_orm(user)
    except AssertionError as e:
        raise HTTPException(status_code=404, detail=str(e))
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


async def list_user(
    db: Session,
    organization_id: int,
    options: UserSearchOptions,
    sortOptions: SortOptions,
):
    def user_transformer(items):
        users = []
        for item in items:
            user, role_id = item
            user.__dict__["role_id"] = role_id
            users.append(user)
        return users

    try:
        sa_sort_list = sortOptions.get_sa_sort_list(
            sa_table=UserModel,
            mapping={"role_id": UserRoleModel.role_id, "role": RoleModel.role},
        )
        users = paginate(
            db,
            query=select(UserModel, UserRoleModel.role_id)
            .join(UserRoleModel, UserRoleModel.user_id == UserModel.id, isouter=True)
            .join(RoleModel, RoleModel.id == UserRoleModel.role_id, isouter=True)
            .filter(
                UserModel.active.is_(True),
                UserModel.organization_id == organization_id,
            )
            .filter(
                UserModel.name.like("%{0}%".format(options.name)),
                UserModel.email.like("%{0}%".format(options.email)),
            )
            .order_by(*sa_sort_list),
            transformer=user_transformer,
            unique=False,
        )
        return users
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


async def create_user(db: Session, data: UserCreate, organization_id: int):
    try:
        user: UserRecord = firebase_auth.create_user(
            email=data.email, password=data.password
        )
        user_db = UserModel(
            uid=user.uid,
            email=user.email,
            name=data.name,
            organization_id=organization_id,
            active=True,
        )
        # it may be possible to specify the organization_id externally in the future
        db.add(user_db)
        db.flush()
        await set_role(db, user_id=user_db.id, role_id=data.role_id, auto_commit=False)
        db.commit()
        user_db.__dict__["role_id"] = data.role_id
        return User.from_orm(user_db)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


async def update_user(
    db: Session, user_id: int, data: UserUpdate, organization_id: int
):
    try:
        user_db = (
            db.query(UserModel)
            .filter(
                UserModel.active.is_(True),
                UserModel.id == user_id,
                UserModel.organization_id == organization_id,
            )
            .first()
        )
        assert user_db is not None, "User not found"
        user_data = data.dict(exclude_unset=True)
        role_id = user_data.pop("role_id", None)
        for key, value in user_data.items():
            setattr(user_db, key, value)
        if role_id is not None:
            await set_role(db, user_id=user_db.id, role_id=role_id, auto_commit=False)
        user_db.__dict__["role_id"] = role_id
        firebase_auth.update_user(user_db.uid, email=data.email)
        db.commit()
        return User.from_orm(user_db)
    except AssertionError as e:
        raise HTTPException(status_code=404, detail=str(e))
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


async def update_password(
    db: Session,
    user_id: int,
    data: UserPasswordUpdate,
    organization_id: int,
):
    user = await get_user(db, user_id, organization_id)
    await authenticate_user(
        db, data=UserAuth(email=user.email, password=data.old_password)
    )
    try:
        user = firebase_auth.update_user(user.uid, password=data.new_password)
        return True
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


async def delete_user(db: Session, user_id: int, organization_id: int) -> bool:
    try:
        user_db = (
            db.query(UserModel)
            .filter(
                UserModel.active.is_(True),
                UserModel.id == user_id,
                UserModel.organization_id == organization_id,
            )
            .first()
        )
        assert user_db is not None, "User not found"
        user_db.active = False
        db.commit()
        firebase_auth.delete_user(user_db.uid)
        return True
    except AssertionError as e:
        raise HTTPException(status_code=404, detail=str(e))
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))



from studio.app.optinist.wrappers.optinist.dimension_reduction.cca import CCA
from studio.app.optinist.wrappers.optinist.dimension_reduction.dpca_fit import dpca_fit
from studio.app.optinist.wrappers.optinist.dimension_reduction.pca import PCA
from studio.app.optinist.wrappers.optinist.dimension_reduction.tsne import TSNE

dimension_reduction_wrapper_dict = {
    "cca": {
        "function": CCA,
        "conda_name": "optinist",
    },
    "dpca": {
        "function": dpca_fit,
        "conda_name": "optinist",
    },
    "pca": {
        "function": PCA,
        "conda_name": "optinist",
    },
    "tsne": {
        "function": TSNE,
        "conda_name": "optinist",
    },
}



import os

import yaml
from fastapi import APIRouter, Depends, File, HTTPException, UploadFile
from fastapi.responses import FileResponse

from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.workflow.workflow import WorkflowConfig
from studio.app.common.core.workflow.workflow_reader import WorkflowConfigReader
from studio.app.common.core.workspace.workspace_dependencies import (
    is_workspace_available,
)
from studio.app.dir_path import DIRPATH

router = APIRouter(prefix="/workflow", tags=["workflow"])


@router.get(
    "/reproduce/{workspace_id}/{unique_id}",
    response_model=WorkflowConfig,
    dependencies=[Depends(is_workspace_available)],
)
async def reproduce_experiment(workspace_id: str, unique_id: str):
    config_filepath = join_filepath(
        [DIRPATH.OUTPUT_DIR, workspace_id, unique_id, DIRPATH.WORKFLOW_YML]
    )
    if os.path.exists(config_filepath):
        return WorkflowConfigReader.read(config_filepath)
    else:
        raise HTTPException(status_code=404, detail="file not found")


@router.get(
    "/download/{workspace_id}/{unique_id}",
    dependencies=[Depends(is_workspace_available)],
)
async def download_workspace_config(workspace_id: str, unique_id: str):
    config_filepath = join_filepath(
        [DIRPATH.OUTPUT_DIR, workspace_id, unique_id, DIRPATH.WORKFLOW_YML]
    )
    if os.path.exists(config_filepath):
        return FileResponse(config_filepath, filename=DIRPATH.WORKFLOW_YML)
    else:
        raise HTTPException(status_code=404, detail="file not found")


@router.post("/import")
async def import_workflow_config(file: UploadFile = File(...)):
    try:
        contents = yaml.safe_load(await file.read())
        if contents is None:
            raise HTTPException(status_code=400, detail="Invalid yaml file")
        return WorkflowConfig(**contents)
    except Exception as e:
        raise HTTPException(status_code=400, detail=f"Parsing yaml failed: {str(e)}")



import argparse
import logging

import uvicorn
from fastapi import Depends, FastAPI, Request
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi_pagination import add_pagination
from starlette.middleware.cors import CORSMiddleware

from studio.app.common.core.auth.auth_dependencies import (
    get_admin_user,
    get_current_user,
)
from studio.app.common.core.mode import MODE
from studio.app.common.core.workspace.workspace_dependencies import (
    is_workspace_available,
    is_workspace_owner,
)
from studio.app.common.routers import (
    algolist,
    auth,
    experiment,
    files,
    outputs,
    params,
    run,
    users_admin,
    users_me,
    users_search,
    workflow,
    workspace,
)
from studio.app.dir_path import DIRPATH
from studio.app.optinist.routers import hdf5, nwb, roi

app = FastAPI(docs_url="/docs", openapi_url="/openapi")

add_pagination(app)

# common routers
app.include_router(algolist.router, dependencies=[Depends(get_current_user)])
app.include_router(auth.router)
app.include_router(experiment.router, dependencies=[Depends(get_current_user)])
app.include_router(files.router, dependencies=[Depends(get_current_user)])
app.include_router(outputs.router, dependencies=[Depends(get_current_user)])
app.include_router(params.router, dependencies=[Depends(get_current_user)])
app.include_router(run.router, dependencies=[Depends(get_current_user)])
app.include_router(users_admin.router, dependencies=[Depends(get_admin_user)])
app.include_router(users_me.router, dependencies=[Depends(get_current_user)])
app.include_router(users_search.router, dependencies=[Depends(get_current_user)])
app.include_router(workflow.router, dependencies=[Depends(get_current_user)])
app.include_router(workspace.router, dependencies=[Depends(get_current_user)])

# optinist routers
app.include_router(hdf5.router, dependencies=[Depends(get_current_user)])
app.include_router(nwb.router, dependencies=[Depends(get_current_user)])
app.include_router(roi.router, dependencies=[Depends(get_current_user)])


def skip_dependencies():
    pass


if MODE.IS_STANDALONE:
    app.dependency_overrides[get_current_user] = skip_dependencies
    app.dependency_overrides[get_admin_user] = skip_dependencies
    app.dependency_overrides[is_workspace_owner] = skip_dependencies
    app.dependency_overrides[is_workspace_available] = skip_dependencies

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

FRONTEND_DIRPATH = DIRPATH.ROOT_DIR + "/frontend"

app.mount(
    "/static",
    StaticFiles(directory=f"{FRONTEND_DIRPATH}/build/static"),
    name="static",
)

templates = Jinja2Templates(directory=f"{FRONTEND_DIRPATH}/build")


@app.on_event("startup")
async def startup_event():
    logging.info('"Studio" application startup complete.')


@app.get("/")
async def root(request: Request):
    return templates.TemplateResponse("index.html", {"request": request})


@app.get("/{_:path}")
async def index(request: Request):
    return await root(request)


def main(develop_mode: bool = False):
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="127.0.0.1")
    parser.add_argument("--port", type=int, default=8000)
    parser.add_argument("--reload", action="store_true")
    args = parser.parse_args()

    log_config_file = (
        f"{DIRPATH.CONFIG_DIR}/standalone-logging.yaml"
        if MODE.IS_STANDALONE
        else f"{DIRPATH.CONFIG_DIR}/logging.yaml"
    )

    if develop_mode:
        reload_options = {"reload_dirs": ["studio"]} if args.reload else {}
        uvicorn.run(
            "studio.__main_unit__:app",
            host=args.host,
            port=args.port,
            log_config=log_config_file,
            reload=args.reload,
            **reload_options,
        )
    else:
        uvicorn.run(
            "studio.__main_unit__:app",
            host=args.host,
            port=args.port,
            log_config=log_config_file,
            reload=False,
        )



from dataclasses import dataclass
from typing import Dict, List, Union

from pydantic import BaseModel


@dataclass
class Arg:
    name: str
    type: str
    isNone: bool


@dataclass
class Return:
    name: str
    type: str


@dataclass
class Algo:
    args: List[Arg]
    returns: List[Return]
    parameter: str = None
    path: str = None


class AlgoModel(BaseModel):
    children: Union[Dict[str, Algo], Dict[str, "AlgoModel"]]


class AlgoList(BaseModel):
    __root__: Dict[str, AlgoModel] = {
        "caiman": {
            "children": {
                "caiman_mc": {
                    "args": [{"name": "image", "type": "ImageData", "isNone": False}],
                    "returns": [{"name": "mc_images", "type": "ImageData"}],
                    "parameter": None,
                    "path": "caiman/caiman_mc",
                }
            }
        }
    }



import json
import logging

import pyrebase

from studio.app.common.core.mode import MODE
from studio.app.dir_path import DIRPATH

try:
    pyrebase_app = pyrebase.initialize_app(
        json.load(open(DIRPATH.FIREBASE_CONFIG_PATH))
    )
except FileNotFoundError as e:
    if MODE.IS_STANDALONE:
        pyrebase_app = None
    else:
        logging.getLogger().error("Firebase config file not found.")
        raise e
except Exception as e:
    logging.getLogger().error(e)
    raise e



import gc
from typing import Optional

from studio.app.common.core.workflow.workflow import OutputPath


class BaseData:
    def __init__(self, file_name):
        self.file_name = file_name

    def save_json(self, json_dir):
        pass

    @property
    def output_path(self) -> Optional[OutputPath]:
        return None

    def __del__(self):
        del self
        gc.collect()



# from studio.app.common.wrappers import wrapper_dict as common_wrapper_dict
from studio.app.optinist.wrappers import wrapper_dict as optinist_wrapper_dict

wrapper_dict = {}
# wrapper_dict.update(**common_wrapper_dict)
wrapper_dict.update(**optinist_wrapper_dict)



from glob import glob

from fastapi import APIRouter, Depends
from fastapi.responses import FileResponse

from studio.app.common.core.utils.config_handler import ConfigReader
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.utils.filepath_finder import find_param_filepath
from studio.app.common.core.workspace.workspace_dependencies import (
    is_workspace_available,
)
from studio.app.dir_path import DIRPATH
from studio.app.optinist.schemas.nwb import NWBParams

router = APIRouter()


@router.get("/nwb", response_model=NWBParams, tags=["params"])
async def get_nwb_params():
    filepath = find_param_filepath("nwb")
    return ConfigReader.read(filepath)


@router.get(
    "/experiments/download/nwb/{workspace_id}/{unique_id}",
    dependencies=[Depends(is_workspace_available)],
    tags=["experiments"],
)
async def download_nwb_experiment(workspace_id: str, unique_id: str):
    nwb_path_list = glob(
        join_filepath([DIRPATH.OUTPUT_DIR, workspace_id, unique_id, "*.nwb"])
    )
    if len(nwb_path_list) > 0:
        return FileResponse(nwb_path_list[0])
    else:
        return False


@router.get(
    "/experiments/download/nwb/{workspace_id}/{unique_id}/{function_id}",
    dependencies=[Depends(is_workspace_available)],
    tags=["experiments"],
)
async def download_nwb_experiment_with_function_id(
    workspace_id: str, unique_id: str, function_id: str
):
    nwb_path_list = glob(
        join_filepath(
            [DIRPATH.OUTPUT_DIR, workspace_id, unique_id, function_id, "*.nwb"]
        )
    )
    if len(nwb_path_list) > 0:
        return FileResponse(nwb_path_list[0])
    else:
        return False



from enum import Enum
from typing import Dict, List, Union

from fastapi import Query
from pydantic import dataclasses
from sqlmodel.main import SQLModelMetaclass


class SortDirection(str, Enum):
    asc = "asc"
    desc = "desc"


@dataclasses.dataclass
class SortOptions:
    sort: List = Query(
        default=(None, SortDirection.asc),
        description="field-0: sort column<br/>field-1: order ('asc' or 'desc')",
    )

    def get_sa_sort_list(
        self, sa_table, mapping: Dict[str, Union[str, SQLModelMetaclass]] = None
    ) -> List:
        sort_list = []
        for i in range(0, len(self.sort), 2):
            sort_field, sort_type = self.sort[i] or "id", self.sort[i + 1]

            sort_column = (
                mapping.get(sort_field) or sort_field if mapping else sort_field
            )
            if isinstance(sort_column, str):
                sort_column = getattr(sa_table, sort_column)

            sort_list.append(
                sort_column.desc()
                if sort_type and sort_type == SortDirection.desc
                else sort_column.asc()
            )
        return sort_list



from typing import List

from fastapi import APIRouter, Depends, Query
from sqlmodel import Session, or_

from studio.app.common.core.auth.auth_dependencies import get_current_user
from studio.app.common.db.database import get_db
from studio.app.common.models import User as UserModel
from studio.app.common.schemas.users import User, UserInfo

router = APIRouter(prefix="/users/search", tags=["users/search"])


@router.get(
    "/share_users",
    response_model=List[UserInfo],
    description="""
- Get a list of users with whom content is shared.
- Note: Maximum of 10 responses. (security considerations)
""",
)
def search_share_users(
    keyword: str = Query(description="partial match (user.name or user.email)"),
    db: Session = Depends(get_db),
    current_user: User = Depends(get_current_user),
):
    MAX_RESPONSE_COUNT = 10
    users = (
        db.query(UserModel)
        .filter(
            UserModel.active.is_(True),
            UserModel.organization_id == current_user.organization.id,
        )
        .filter(
            or_(
                UserModel.name.like("%{0}%".format(keyword)),
                UserModel.email.like("%{0}%".format(keyword)),
            )
        )
        .order_by(UserModel.id)
        .limit(MAX_RESPONSE_COUNT)
        .all()
    )
    return users



from fastapi import APIRouter, Depends
from sqlmodel import Session

from studio.app.common.core.auth.auth_dependencies import get_current_user
from studio.app.common.core.users import crud_users
from studio.app.common.db.database import get_db
from studio.app.common.schemas.users import SelfUserUpdate, User, UserPasswordUpdate

router = APIRouter(prefix="/users/me", tags=["users/me"])


@router.get("", response_model=User)
async def me(current_user: User = Depends(get_current_user)):
    return current_user


@router.put("", response_model=User)
async def update_me(
    data: SelfUserUpdate,
    current_user: User = Depends(get_current_user),
    db: Session = Depends(get_db),
):
    return await crud_users.update_user(
        db, current_user.id, data, organization_id=current_user.organization.id
    )


@router.put("/password", response_model=bool)
async def update_password(
    data: UserPasswordUpdate,
    current_user: User = Depends(get_current_user),
    db: Session = Depends(get_db),
):
    return await crud_users.update_password(
        db, current_user.id, data, organization_id=current_user.organization.id
    )


@router.delete("", response_model=bool)
async def delete_me(
    current_user: User = Depends(get_current_user),
    db: Session = Depends(get_db),
):
    return await crud_users.delete_user(
        db, current_user.id, organization_id=current_user.organization.id
    )



from studio.app.common.core.workflow.workflow import WorkflowConfig


class WorkflowConfigBuilder:
    def __init__(self):
        self._nodeDict = None
        self._edgeDict = None

    def set_config(self, config: WorkflowConfig) -> "WorkflowConfigBuilder":
        self._nodeDict = config.nodeDict
        self._edgeDict = config.edgeDict
        return self

    def set_node_dict(self, nodeDict) -> "WorkflowConfigBuilder":
        self._nodeDict = nodeDict
        return self

    def set_edge_dict(self, edgeDict) -> "WorkflowConfigBuilder":
        self._edgeDict = edgeDict
        return self

    def build(self) -> WorkflowConfig:
        return WorkflowConfig(self._nodeDict, self._edgeDict)



from dataclasses import dataclass
from typing import Dict, List, Union


@dataclass
class OutputData:
    data: Union[List, Dict, str]
    columns: List[str] = None
    index: List[str] = None


@dataclass
class JsonTimeSeriesData(OutputData):
    xrange: list = None
    std: Dict[str, dict] = None



import json
import logging

from fastapi import HTTPException, status
from fastapi.responses import JSONResponse
from requests.exceptions import HTTPError
from sqlmodel import Session

from studio.app.common.core.auth import pyrebase_app
from studio.app.common.core.auth.security import (
    create_access_token,
    create_refresh_token,
    validate_refresh_token,
)
from studio.app.common.models.user import User as UserModel
from studio.app.common.schemas.auth import AccessToken, Token, UserAuth


async def authenticate_user(db: Session, data: UserAuth):
    try:
        user = pyrebase_app.auth().sign_in_with_email_and_password(
            data.email, data.password
        )
        user_db = (
            db.query(UserModel)
            .filter(UserModel.uid == user["localId"], UserModel.active.is_(True))
            .first()
        )
        assert user_db is not None, "Invalid user uid"
        ex_token = create_access_token(subject=user_db.uid)
        return Token(
            access_token=user["idToken"],
            refresh_token=create_refresh_token(subject=user["refreshToken"]),
            token_type="bearer",
            ex_token=ex_token,
        )
    except Exception as e:
        logging.getLogger().error(e)
        raise HTTPException(status_code=400, detail=str(e))


async def refresh_current_user_token(refresh_token: str):
    token, err = validate_refresh_token(refresh_token)

    if err:
        raise HTTPException(status_code=400)
    try:
        user = pyrebase_app.auth().refresh(refresh_token=token["sub"])
        return AccessToken(access_token=user["idToken"])
    except Exception as e:
        logging.getLogger().error(e)
        raise HTTPException(status_code=400)


async def send_reset_password_mail(db: Session, email: str):
    try:
        db.query(UserModel).filter(
            UserModel.email == email, UserModel.active.is_(True)
        ).one()
        pyrebase_app.auth().send_password_reset_email(email)
        return JSONResponse(content=None, status_code=status.HTTP_200_OK)
    except HTTPError as e:
        logging.getLogger().error(e)
        err = json.loads(e.strerror)
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=err.get("error").get("message"),
        )
    except Exception as e:
        logging.getLogger().error(e)
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)



from firebase_admin import credentials, initialize_app

from studio.app.dir_path import DIRPATH

try:
    initialize_app(credentials.Certificate(DIRPATH.FIREBASE_PRIVATE_PATH))
except Exception:
    pass



from dataclasses import asdict
from typing import Dict

from studio.app.common.core.utils.config_handler import ConfigWriter
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.workflow.workflow import Edge, Node, WorkflowConfig
from studio.app.common.core.workflow.workflow_builder import WorkflowConfigBuilder
from studio.app.dir_path import DIRPATH


class WorkflowConfigWriter:
    def __init__(
        self,
        workspace_id: str,
        unique_id: str,
        nodeDict: Dict[str, Node],
        edgeDict: Dict[str, Edge],
    ) -> None:
        self.workspace_id = workspace_id
        self.unique_id = unique_id
        self.nodeDict = nodeDict
        self.edgeDict = edgeDict
        self.builder = WorkflowConfigBuilder()
        self.create_config()

    def write(self) -> None:
        ConfigWriter.write(
            dirname=join_filepath(
                [DIRPATH.OUTPUT_DIR, self.workspace_id, self.unique_id]
            ),
            filename=DIRPATH.WORKFLOW_YML,
            config=asdict(self.builder.build()),
        )

    def create_config(self) -> WorkflowConfig:
        return (
            self.builder.set_node_dict(self.nodeDict)
            .set_edge_dict(self.edgeDict)
            .build()
        )



from studio.app.common.dataclass.bar import BarData
from studio.app.common.dataclass.base import BaseData
from studio.app.common.dataclass.csv import CsvData
from studio.app.common.dataclass.heatmap import HeatMapData
from studio.app.common.dataclass.histogram import HistogramData
from studio.app.common.dataclass.html import HTMLData
from studio.app.common.dataclass.image import ImageData
from studio.app.common.dataclass.line import LineData
from studio.app.common.dataclass.pie import PieData
from studio.app.common.dataclass.polar import PolarData
from studio.app.common.dataclass.scatter import ScatterData
from studio.app.common.dataclass.timeseries import TimeSeriesData

__all__ = [
    "BarData",
    "BaseData",
    "CsvData",
    "HeatMapData",
    "HistogramData",
    "HTMLData",
    "ImageData",
    "LineData",
    "PieData",
    "PolarData",
    "ScatterData",
    "TimeSeriesData",
]



from dataclasses import dataclass


@dataclass
class FILETYPE:
    IMAGE: str = "image"
    CSV: str = "csv"
    HDF5: str = "hdf5"
    BEHAVIOR: str = "behavior"


ACCEPT_TIFF_EXT = [".tif", ".tiff", ".TIF", ".TIFF"]
ACCEPT_CSV_EXT = [".csv"]
ACCEPT_HDF5_EXT = [".hdf5", ".nwb", ".HDF5", ".NWB"]

NOT_DISPLAY_ARGS_LIST = ["params", "output_dir", "nwbfile", "kwargs"]

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"



from typing import Any, Dict

from fastapi import APIRouter

from studio.app.common.core.utils.config_handler import ConfigReader
from studio.app.common.core.utils.filepath_finder import find_param_filepath
from studio.app.common.schemas.params import SnakemakeParams

router = APIRouter(tags=["params"])


@router.get("/params/{name}", response_model=Dict[str, Any])
async def get_params(name: str):
    filepath = find_param_filepath(name)
    config = ConfigReader.read(filepath)
    return config


@router.get("/snakemake", response_model=SnakemakeParams)
async def get_snakemake_params():
    filepath = find_param_filepath("snakemake")
    return ConfigReader.read(filepath)



from studio.app.common.core.utils.config_handler import ConfigWriter
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.dir_path import DIRPATH


class SmkConfigWriter:
    @classmethod
    def write(cls, workspace_id, unique_id, flow_config):
        ConfigWriter.write(
            dirname=join_filepath([DIRPATH.OUTPUT_DIR, workspace_id, unique_id]),
            filename=DIRPATH.SNAKEMAKE_CONFIG_YML,
            config=flow_config,
        )



import os

import yaml

from studio.app.common.core.utils.filepath_creater import (
    create_directory,
    join_filepath,
)


class ConfigReader:
    @classmethod
    def read(cls, filepath):
        config = {}
        if filepath is not None and os.path.exists(filepath):
            with open(filepath) as f:
                config = yaml.safe_load(f)
        return config


class ConfigWriter:
    @classmethod
    def write(cls, dirname, filename, config):
        create_directory(dirname)

        with open(join_filepath([dirname, filename]), "w") as f:
            yaml.dump(config, f)



from fastapi import APIRouter, Depends

from studio.app.common.core.workspace.workspace_dependencies import is_workspace_owner
from studio.app.optinist.core.edit_ROI import ACTION, EditROI
from studio.app.optinist.schemas.roi import EditRoiSuccess, RoiList, RoiPos

router = APIRouter(prefix="/outputs", tags=["outputs"])


@router.post(
    "/image/{filepath:path}/add_roi",
    response_model=EditRoiSuccess,
    dependencies=[Depends(is_workspace_owner)],
)
async def add_roi(filepath: str, pos: RoiPos):
    EditROI(action=ACTION.ADD, filepath=filepath, params=pos.dict()).excute()
    return EditRoiSuccess(max_index=0)


@router.post(
    "/image/{filepath:path}/merge_roi",
    response_model=EditRoiSuccess,
    dependencies=[Depends(is_workspace_owner)],
)
async def merge_roi(filepath: str, roi_list: RoiList):
    EditROI(action=ACTION.MERGE, filepath=filepath, params=roi_list.dict()).excute()
    return EditRoiSuccess(max_index=0)


@router.post(
    "/image/{filepath:path}/delete_roi",
    response_model=EditRoiSuccess,
    dependencies=[Depends(is_workspace_owner)],
)
async def delete_roi(filepath: str, roi_list: RoiList):
    EditROI(action=ACTION.DELETE, filepath=filepath, params=roi_list.dict()).excute()
    return EditRoiSuccess(max_index=0)



from typing import List

from pydantic import BaseModel, Field


class RoiPos(BaseModel):
    posx: int
    posy: int
    sizex: int
    sizey: int


class RoiList(BaseModel):
    ids: List[int] = Field(default=[0, 1])


class EditRoiSuccess(BaseModel):
    max_index: int



from typing import Any, Dict, Optional, Union

from pydantic import BaseModel


class NWBParams(BaseModel):
    session_description: str = "optinist"
    identifier: str = "optinist"
    experiment_description: Optional[str] = None
    device: Union[Dict, Any]
    optical_channel: Union[Dict, Any]
    imaging_plane: Union[Dict, Any]
    image_series: Union[Dict, Any]
    ophys: Union[Dict, Any]



from dataclasses import dataclass
from typing import Dict, List, Optional

from pydantic import BaseModel

from studio.app.common.core.experiment.experiment import ExptFunction
from studio.app.common.core.workflow.workflow import Edge, Node


class DeleteItem(BaseModel):
    uidList: List


class RenameItem(BaseModel):
    new_name: str


@dataclass
class FetchExptResponse:
    workspace_id: str
    unique_id: str
    name: str
    started_at: str
    finished_at: Optional[str]
    success: Optional[str]
    hasNWB: bool
    function: Dict[str, ExptFunction]
    nodeDict: Dict[str, Node]
    edgeDict: Dict[str, Edge]



from dataclasses import dataclass
from typing import Dict, Optional

from studio.app.common.core.snakemake.smk import SmkParam
from studio.app.common.core.workflow.workflow import OutputPath
from studio.app.optinist.schemas.nwb import NWBParams


@dataclass
class ExptFunction:
    unique_id: str
    name: str
    success: str
    hasNWB: bool
    message: Optional[str] = None
    outputPaths: Optional[Dict[str, OutputPath]] = None
    started_at: Optional[str] = None
    finished_at: Optional[str] = None


@dataclass
class ExptConfig:
    workspace_id: str
    unique_id: str
    name: str
    started_at: str
    finished_at: Optional[str]
    success: Optional[str]
    hasNWB: bool
    function: Dict[str, ExptFunction]
    nwb: NWBParams
    snakemake: SmkParam



from typing import List

from pydantic.dataclasses import dataclass as pydantic_dataclass


@pydantic_dataclass
class HDF5Node:
    isDir: bool
    name: str
    path: str
    nodes: List["HDF5Node"] = None
    shape: tuple = None
    nbytes: str = None



from pydantic import BaseSettings, Field

from studio.app.dir_path import DIRPATH


class Mode(BaseSettings):
    IS_STANDALONE: bool = Field(default=True, env="IS_STANDALONE")

    class Config:
        env_file = f"{DIRPATH.CONFIG_DIR}/.env"
        env_file_encoding = "utf-8"


MODE = Mode()



import os
from dataclasses import asdict
from datetime import datetime
from typing import Dict

from studio.app.common.core.experiment.experiment import ExptConfig, ExptFunction
from studio.app.common.core.experiment.experiment_builder import ExptConfigBuilder
from studio.app.common.core.experiment.experiment_reader import ExptConfigReader
from studio.app.common.core.utils.config_handler import ConfigWriter
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.workflow.workflow_reader import WorkflowConfigReader
from studio.app.const import DATE_FORMAT
from studio.app.dir_path import DIRPATH


class ExptConfigWriter:
    def __init__(
        self,
        workspace_id: str,
        unique_id: str,
        name: str,
        nwbfile: Dict = {},
        snakemake: Dict = {},
    ) -> None:
        self.workspace_id = workspace_id
        self.unique_id = unique_id
        self.name = name
        self.nwbfile = nwbfile
        self.snakemake = snakemake
        self.builder = ExptConfigBuilder()

    def write(self) -> None:
        expt_filepath = join_filepath(
            [
                DIRPATH.OUTPUT_DIR,
                self.workspace_id,
                self.unique_id,
                DIRPATH.EXPERIMENT_YML,
            ]
        )
        if os.path.exists(expt_filepath):
            expt_config = ExptConfigReader.read(expt_filepath)
            self.builder.set_config(expt_config)
            self.add_run_info()
        else:
            self.create_config()

        self.function_from_nodeDict()

        ConfigWriter.write(
            dirname=join_filepath(
                [DIRPATH.OUTPUT_DIR, self.workspace_id, self.unique_id]
            ),
            filename=DIRPATH.EXPERIMENT_YML,
            config=asdict(self.builder.build()),
        )

    def create_config(self) -> ExptConfig:
        return (
            self.builder.set_workspace_id(self.workspace_id)
            .set_unique_id(self.unique_id)
            .set_name(self.name)
            .set_started_at(datetime.now().strftime(DATE_FORMAT))
            .set_success("running")
            .set_nwbfile(self.nwbfile)
            .set_snakemake(self.snakemake)
            .build()
        )

    def add_run_info(self) -> ExptConfig:
        return (
            self.builder.set_started_at(datetime.now().strftime(DATE_FORMAT))  # 時間を更新
            .set_success("running")
            .build()
        )

    def function_from_nodeDict(self) -> ExptConfig:
        func_dict: Dict[str, ExptFunction] = {}
        node_dict = WorkflowConfigReader.read(
            join_filepath(
                [
                    DIRPATH.OUTPUT_DIR,
                    self.workspace_id,
                    self.unique_id,
                    DIRPATH.WORKFLOW_YML,
                ]
            )
        ).nodeDict

        for node in node_dict.values():
            func_dict[node.id] = ExptFunction(
                unique_id=node.id, name=node.data.label, hasNWB=False, success="running"
            )
            if node.data.type == "input":
                timestamp = datetime.now().strftime(DATE_FORMAT)
                func_dict[node.id].started_at = timestamp
                func_dict[node.id].finished_at = timestamp
                func_dict[node.id].success = "success"

        return self.builder.set_function(func_dict).build()



from studio.app.optinist.core.edit_ROI.edit_ROI import ACTION, EditROI, EditRoiUtils
from studio.app.optinist.core.edit_ROI.wrappers import edit_roi_wrapper_dict

__all__ = ["ACTION", "EditROI", "EditRoiUtils", "edit_roi_wrapper_dict"]



from fastapi import APIRouter, Depends
from fastapi_pagination import LimitOffsetPage
from sqlmodel import Session

from studio.app.common.core.auth.auth_dependencies import get_admin_user
from studio.app.common.core.users import crud_users
from studio.app.common.db.database import get_db
from studio.app.common.schemas.base import SortOptions
from studio.app.common.schemas.users import (
    User,
    UserCreate,
    UserSearchOptions,
    UserUpdate,
)

router = APIRouter(prefix="/admin/users", tags=["users/admin"])


@router.get("", response_model=LimitOffsetPage[User])
async def list_user(
    db: Session = Depends(get_db),
    options: UserSearchOptions = Depends(),
    sortOptions: SortOptions = Depends(),
    current_admin: User = Depends(get_admin_user),
):
    return await crud_users.list_user(
        db,
        organization_id=current_admin.organization.id,
        options=options,
        sortOptions=sortOptions,
    )


@router.post("", response_model=User)
async def create_user(
    data: UserCreate,
    db: Session = Depends(get_db),
    current_admin: User = Depends(get_admin_user),
):
    return await crud_users.create_user(
        db, data, organization_id=current_admin.organization.id
    )


@router.get("/{user_id}", response_model=User)
async def get_user(
    user_id: int,
    db: Session = Depends(get_db),
    current_admin: User = Depends(get_admin_user),
):
    return await crud_users.get_user(
        db, user_id, organization_id=current_admin.organization.id
    )


@router.put("/{user_id}", response_model=User)
async def update_user(
    user_id: int,
    data: UserUpdate,
    db: Session = Depends(get_db),
    current_admin: User = Depends(get_admin_user),
):
    return await crud_users.update_user(
        db, user_id, data, organization_id=current_admin.organization.id
    )


@router.delete("/{user_id}", response_model=bool)
async def delete_user(
    user_id: int,
    db: Session = Depends(get_db),
    current_admin: User = Depends(get_admin_user),
):
    return await crud_users.delete_user(
        db, user_id, organization_id=current_admin.organization.id
    )



from dataclasses import dataclass
from typing import List

from pydantic.dataclasses import dataclass as pydantic_dataclass


@pydantic_dataclass
class TreeNode:
    path: str
    name: str
    isdir: bool
    nodes: List["TreeNode"]


@dataclass
class FilePath:
    file_path: str



import os
from glob import glob
from typing import Optional

from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.dir_path import CORE_PARAM_PATH, DIRPATH


def find_filepath(name, category) -> Optional[str]:
    name, _ = os.path.splitext(name)

    filepaths = glob(
        join_filepath(
            [DIRPATH.APP_DIR, "*", "wrappers", "**", category, f"{name}.yaml"]
        ),
        recursive=True,
    )
    return filepaths[0] if len(filepaths) > 0 else None


def find_param_filepath(name: str):
    if name in CORE_PARAM_PATH.__members__:
        return CORE_PARAM_PATH[name].value
    else:
        return find_filepath(name, "params")


def find_condaenv_filepath(name: str):
    return find_filepath(name, "conda")



"""add common tables

Revision ID: a19721f73cfb
Revises:
Create Date: 2023-09-21 18:02:50.136222

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "a19721f73cfb"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "organization",
        sa.Column("id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("CURRENT_TIMESTAMP"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "roles",
        sa.Column("id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("role", sa.String(length=100), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("CURRENT_TIMESTAMP"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "users",
        sa.Column("id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("organization_id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("uid", sa.String(length=100), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column("email", sa.String(length=255), nullable=False),
        sa.Column("attributes", sa.JSON(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("CURRENT_TIMESTAMP"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(),
            server_default=sa.text("CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP"),
            nullable=True,
        ),
        sa.Column("active", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["organization_id"],
            ["organization.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("uid", name="idx_uid"),
    )
    op.create_table(
        "user_roles",
        sa.Column("id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("user_id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("role_id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("CURRENT_TIMESTAMP"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["role_id"],
            ["roles.id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["users.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "role_id", name="idx_user_id"),
    )
    op.create_table(
        "workspaces",
        sa.Column("id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column("user_id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("CURRENT_TIMESTAMP"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(),
            server_default=sa.text("CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP"),
            nullable=True,
        ),
        sa.Column("deleted", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(["user_id"], ["users.id"], name="user"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "workspaces_share_users",
        sa.Column("id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("workspace_id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column("user_id", mysql.BIGINT(unsigned=True), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            server_default=sa.text("CURRENT_TIMESTAMP"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["users.id"],
        ),
        sa.ForeignKeyConstraint(
            ["workspace_id"],
            ["workspaces.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("workspace_id", "user_id", name="idx_workspace_id_user_id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("workspaces_share_users")
    op.drop_table("workspaces")
    op.drop_table("user_roles")
    op.drop_table("users")
    op.drop_table("roles")
    op.drop_table("organization")
    # ### end Alembic commands ###



from studio.app.optinist.wrappers.caiman.cnmf import caiman_cnmf
from studio.app.optinist.wrappers.caiman.cnmfe import caiman_cnmfe
from studio.app.optinist.wrappers.caiman.motion_correction import caiman_mc

caiman_wrapper_dict = {
    "caiman": {
        "caiman_mc": {
            "function": caiman_mc,
            "conda_name": "caiman",
        },
        "caiman_cnmf": {
            "function": caiman_cnmf,
            "conda_name": "caiman",
        },
        "caiman_cnmfe": {
            "function": caiman_cnmfe,
            "conda_name": "caiman",
        },
    }
}



import uuid
from typing import Dict

from fastapi import APIRouter, BackgroundTasks, Depends

from studio.app.common.core.workflow.workflow import Message, NodeItem, RunItem
from studio.app.common.core.workflow.workflow_result import WorkflowResult
from studio.app.common.core.workflow.workflow_runner import WorkflowRunner
from studio.app.common.core.workspace.workspace_dependencies import (
    is_workspace_available,
    is_workspace_owner,
)

router = APIRouter(prefix="/run", tags=["run"])


@router.post(
    "/{workspace_id}",
    response_model=str,
    dependencies=[Depends(is_workspace_owner)],
)
async def run(workspace_id: str, runItem: RunItem, background_tasks: BackgroundTasks):
    unique_id = str(uuid.uuid4())[:8]
    WorkflowRunner(workspace_id, unique_id, runItem).run_workflow(background_tasks)
    print("run snakemake")
    return unique_id


@router.post(
    "/{workspace_id}/{uid}",
    response_model=str,
    dependencies=[Depends(is_workspace_owner)],
)
async def run_id(
    workspace_id: str, uid: str, runItem: RunItem, background_tasks: BackgroundTasks
):
    WorkflowRunner(workspace_id, uid, runItem).run_workflow(background_tasks)
    print("run snakemake")
    print("forcerun list: ", runItem.forceRunList)
    return uid


@router.post(
    "/result/{workspace_id}/{uid}",
    response_model=Dict[str, Message],
    dependencies=[Depends(is_workspace_available)],
)
async def run_result(workspace_id: str, uid: str, nodeDict: NodeItem):
    return WorkflowResult(workspace_id, uid).get(nodeDict.pendingNodeIdList)


@router.post(
    "/cancel/{workspace_id}/{uid}",
    response_model=bool,
    dependencies=[Depends(is_workspace_owner)],
)
async def run_cancel(workspace_id: str, uid: str):
    return WorkflowResult(workspace_id, uid).cancel()



import os
import signal
from dataclasses import asdict
from datetime import datetime
from glob import glob
from typing import Dict

from fastapi import HTTPException

from studio.app.common.core.experiment.experiment_reader import ExptConfigReader
from studio.app.common.core.utils.config_handler import ConfigWriter
from studio.app.common.core.utils.file_reader import JsonReader, Reader
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.utils.pickle_handler import PickleReader
from studio.app.common.core.workflow.workflow import Message, OutputPath
from studio.app.common.dataclass import BaseData
from studio.app.const import DATE_FORMAT
from studio.app.dir_path import DIRPATH


class WorkflowResult:
    def __init__(self, workspace_id, unique_id):
        self.workspace_id = workspace_id
        self.unique_id = unique_id
        self.workflow_dirpath = join_filepath(
            [
                DIRPATH.OUTPUT_DIR,
                self.workspace_id,
                self.unique_id,
            ]
        )
        self.expt_filepath = join_filepath(
            [self.workflow_dirpath, DIRPATH.EXPERIMENT_YML]
        )
        self.error_filepath = join_filepath([self.workflow_dirpath, "error.log"])
        self.pid_filepath = join_filepath([self.workflow_dirpath, "pid.json"])

    def get(self, nodeIdList):
        results: Dict[str, Message] = {}
        for node_id in nodeIdList:
            if os.path.exists(self.error_filepath):
                error_message = Reader.read(self.error_filepath)
                if error_message != "":
                    results[node_id] = Message(
                        status="error",
                        message=error_message,
                    )

            glob_pickle_filepath = join_filepath(
                [self.workflow_dirpath, node_id, "*.pkl"]
            )
            for pickle_filepath in glob(glob_pickle_filepath):
                results[node_id] = NodeResult(
                    self.workflow_dirpath,
                    node_id,
                    pickle_filepath,
                ).get()
                self.has_nwb(node_id)

        self.has_nwb()

        return results

    def has_nwb(self, node_id=None):
        if node_id is None:
            nwb_filepath_list = glob(join_filepath([self.workflow_dirpath, "*.nwb"]))
        else:
            nwb_filepath_list = glob(
                join_filepath([self.workflow_dirpath, node_id, "*.nwb"])
            )

        for nwb_filepath in nwb_filepath_list:
            if os.path.exists(nwb_filepath):
                config = ExptConfigReader.read(self.expt_filepath)

                if node_id is None:
                    config.hasNWB = True
                else:
                    config.function[node_id].hasNWB = True

                ConfigWriter.write(
                    dirname=self.workflow_dirpath,
                    filename=DIRPATH.EXPERIMENT_YML,
                    config=asdict(config),
                )

    def cancel(self):
        """
        The algorithm function of this workflow is being executed at the line:
        https://github.com/snakemake/snakemake/blob/27b224ed12448df8aebc7d1ff8f25e3bf7622232/snakemake/shell.py#L258
        ```
        proc = sp.Popen(
            cmd,
            bufsize=-1,
            shell=use_shell,
            stdout=stdout,
            universal_newlines=iterable or read or None,
            close_fds=close_fds,
            **cls._process_args,
            env=envvars,
        )
        ```
        The `cmd` argument has the following format:
        ```
        source ~/miniconda3/bin/activate
        '~/Documents/optinistfs/.snakemake/conda/491889952d2f07f3876bb801eea626e9_';
        set -euo pipefail;
        python ~/Documents/optinistfs/.snakemake/scripts/tmp03froqxo.func.py
        ```
        Interrupt the conda activate at the beginning of the process is impossible
        because it is only called when each algorithm function executes.
        This workflow is cancelled by killing process via PID of algorithm function
        saved in pid.json file
        Raises:
            HTTPException: if pid_filepath or last_script_file does not exist
        """
        if not os.path.exists(self.pid_filepath):
            raise HTTPException(status_code=404)

        pid_data = JsonReader.read(self.pid_filepath)

        if not os.path.exists(pid_data["last_script_file"]):
            raise HTTPException(status_code=404)

        os.remove(pid_data["last_script_file"])
        os.kill(pid_data["last_pid"], signal.SIGTERM)

        return True


class NodeResult:
    def __init__(self, workflow_dirpath, node_id, pickle_filepath):
        self.workflow_dirpath = workflow_dirpath
        self.node_id = node_id
        self.node_dirpath = join_filepath([self.workflow_dirpath, self.node_id])
        self.expt_filepath = join_filepath(
            [self.workflow_dirpath, DIRPATH.EXPERIMENT_YML]
        )

        pickle_filepath = pickle_filepath.replace("\\", "/")
        self.algo_name = os.path.splitext(os.path.basename(pickle_filepath))[0]
        self.info = PickleReader.read(pickle_filepath)

    def get(self):
        expt_config = ExptConfigReader.read(self.expt_filepath)
        if isinstance(self.info, (list, str)):
            expt_config.function[self.node_id].success = "error"
            message = self.error()
        else:
            expt_config.function[self.node_id].success = "success"
            message = self.success()
            expt_config.function[self.node_id].outputPaths = message.outputPaths
        now = datetime.now().strftime(DATE_FORMAT)
        expt_config.function[self.node_id].finished_at = now
        expt_config.function[self.node_id].message = message.message

        statuses = list(map(lambda x: x.success, expt_config.function.values()))
        if "running" not in statuses:
            expt_config.finished_at = now
            if "error" in statuses:
                expt_config.success = "error"
            else:
                expt_config.success = "success"

        ConfigWriter.write(
            dirname=self.workflow_dirpath,
            filename=DIRPATH.EXPERIMENT_YML,
            config=asdict(expt_config),
        )

        return message

    def success(self):
        return Message(
            status="success",
            message=f"{self.algo_name} success",
            outputPaths=self.outputPaths(),
        )

    def error(self):
        return Message(
            status="error",
            message=self.info if isinstance(self.info, str) else "\n".join(self.info),
        )

    def outputPaths(self):
        outputPaths: Dict[str, OutputPath] = {}
        for k, v in self.info.items():
            if isinstance(v, BaseData):
                v.save_json(self.node_dirpath)
                if v.output_path:
                    outputPaths[k] = v.output_path

        return outputPaths



from datetime import datetime
from enum import Enum
from typing import Optional

from fastapi import Query
from pydantic import BaseModel, EmailStr, Field

password_regex = r"^(?=.*\d)(?=.*[!#$%&()*+,-./@_|])(?=.*[a-zA-Z]).{6,255}$"


class Organization(BaseModel):
    id: int
    name: str

    class Config:
        orm_mode = True


class UserSearchOptions(BaseModel):
    email: Optional[str] = Field(Query(default=""))
    name: Optional[str] = Field(Query(default=""))


class UserRole(int, Enum):
    admin = 1
    operator = 20


class User(BaseModel):
    id: int
    uid: str
    name: Optional[str]
    email: EmailStr
    organization: Organization
    role_id: Optional[int]

    @property
    def is_admin(self) -> bool:
        return self.role_id == UserRole.admin

    class Config:
        orm_mode = True


class UserCreate(BaseModel):
    email: EmailStr
    password: str = Field(max_length=255, regex=password_regex)
    name: str
    role_id: int

    class Config:
        anystr_strip_whitespace = True


class UserUpdate(BaseModel):
    email: Optional[EmailStr]
    name: str
    role_id: int


class SelfUserUpdate(BaseModel):
    email: Optional[EmailStr]
    name: str


class UserPasswordUpdate(BaseModel):
    old_password: str
    new_password: str = Field(max_length=255, regex=password_regex)

    class Config:
        anystr_strip_whitespace = True


class UserInfo(BaseModel):
    id: int
    name: Optional[str]
    email: Optional[str]
    created_at: Optional[datetime]
    updated_at: Optional[datetime]

    class Config:
        orm_mode = True



from datetime import datetime
from typing import Dict, List, Optional

from sqlalchemy.dialects.mysql import BIGINT
from sqlalchemy.sql.functions import current_timestamp
from sqlmodel import (
    JSON,
    Column,
    Field,
    ForeignKey,
    Relationship,
    String,
    UniqueConstraint,
)

from studio.app.common.models.base import Base, TimestampMixin
from studio.app.common.models.workspace import WorkspacesShareUser


class UserRole(Base, table=True):
    __tablename__ = "user_roles"
    __table_args__ = (UniqueConstraint("user_id", "role_id", name="idx_user_id"),)

    user_id: int = Field(
        sa_column=Column(BIGINT(unsigned=True), ForeignKey("users.id"), nullable=False),
    )
    role_id: int = Field(
        sa_column=Column(BIGINT(unsigned=True), ForeignKey("roles.id"), nullable=False),
    )
    created_at: Optional[datetime] = Field(
        sa_column_kwargs={"server_default": current_timestamp()},
    )


class User(Base, TimestampMixin, table=True):
    __tablename__ = "users"
    __table_args__ = (UniqueConstraint("uid", name="idx_uid"),)

    organization_id: int = Field(
        sa_column=Column(
            BIGINT(unsigned=True), ForeignKey("organization.id"), nullable=False
        ),
    )
    uid: str = Field(sa_column=Column(String(100), nullable=False))
    name: str = Field(sa_column=Column(String(100), nullable=False))
    email: str = Field(sa_column=Column(String(255), nullable=False))
    attributes: Optional[Dict] = Field(default={}, sa_column=Column(JSON))
    active: bool = Field(nullable=False)

    workspace: List["Workspace"] = Relationship(  # noqa: F821
        back_populates="user", sa_relationship_kwargs={"uselist": True}
    )
    organization: "Organization" = Relationship(
        back_populates="users", sa_relationship_kwargs={"uselist": False}
    )
    role: Optional["Role"] = Relationship(
        back_populates="user",
        link_model=UserRole,
        sa_relationship_kwargs=dict(uselist=False),
    )
    workspace_share: List["Workspace"] = Relationship(  # noqa: F821
        back_populates="user", link_model=WorkspacesShareUser
    )


class Role(Base, table=True):
    __tablename__ = "roles"

    role: str = Field(sa_column=Column(String(100), nullable=False))
    created_at: Optional[datetime] = Field(
        sa_column_kwargs={"server_default": current_timestamp()},
    )
    user: List["User"] = Relationship(back_populates="role", link_model=UserRole)


class Organization(Base, table=True):
    __tablename__ = "organization"

    name: str = Field(sa_column=Column(String(100), nullable=False))
    created_at: Optional[datetime] = Field(
        sa_column_kwargs={"server_default": current_timestamp()},
    )

    users: List["User"] = Relationship(
        back_populates="organization", sa_relationship_kwargs={"uselist": True}
    )



from studio.app.common.dataclass import ImageData
from studio.app.optinist.dataclass import FluoData, IscellData
from studio.app.optinist.wrappers.caiman.cnmf import caiman_cnmf


def caiman_cnmfe(
    images: ImageData, output_dir: str, params: dict = None, **kwargs
) -> dict(fluorescence=FluoData, iscell=IscellData):
    cnmfe_fixed_params = {
        "center_psf": True,
        "method_init": "corr_pnr",  # use this for 1 photon
        "only_init": True,  # set it to True to run CNMF-E
        "normalize_init": False,
    }
    params["fixed_params"] = cnmfe_fixed_params

    return caiman_cnmf(images=images, output_dir=output_dir, params=params, **kwargs)



import os
from collections import deque
from typing import Dict

from snakemake import snakemake

from studio.app.common.core.logger import Logger
from studio.app.common.core.snakemake.smk import SmkParam
from studio.app.common.core.utils.filepath_creater import get_pickle_file, join_filepath
from studio.app.common.core.workflow.workflow import Edge, Node
from studio.app.dir_path import DIRPATH


def snakemake_execute(workspace_id: str, unique_id: str, params: SmkParam):
    logger = Logger(workspace_id, unique_id)
    snakemake(
        DIRPATH.SNAKEMAKE_FILEPATH,
        forceall=params.forceall,
        cores=params.cores,
        use_conda=params.use_conda,
        workdir=f"{os.path.dirname(DIRPATH.STUDIO_DIR)}",
        configfiles=[
            join_filepath(
                [
                    DIRPATH.OUTPUT_DIR,
                    workspace_id,
                    unique_id,
                    DIRPATH.SNAKEMAKE_CONFIG_YML,
                ]
            )
        ],
        log_handler=[logger.smk_logger],
    )
    logger.clean_up()


def delete_dependencies(
    workspace_id: str,
    unique_id: str,
    smk_params: SmkParam,
    nodeDict: Dict[str, Node],
    edgeDict: Dict[str, Edge],
):
    queue = deque()

    for param in smk_params.forcerun:
        queue.append(param.nodeId)

    while True:
        # terminate condition
        if len(queue) == 0:
            break

        # delete pickle
        node_id = queue.pop()
        algo_name = nodeDict[node_id].data.label

        pickle_filepath = join_filepath(
            [
                DIRPATH.OUTPUT_DIR,
                get_pickle_file(
                    workspace_id=workspace_id,
                    unique_id=unique_id,
                    node_id=node_id,
                    algo_name=algo_name,
                ),
            ]
        )
        # print(pickle_filepath)

        if os.path.exists(pickle_filepath):
            os.remove(pickle_filepath)

        # 全てのedgeを見て、node_idがsourceならtargetをqueueに追加する
        for edge in edgeDict.values():
            if node_id == edge.source:
                queue.append(edge.target)



import os
import shutil
from glob import glob
from typing import List

from fastapi import APIRouter, Depends, File, UploadFile

from studio.app.common.core.utils.filepath_creater import (
    create_directory,
    join_filepath,
)
from studio.app.common.core.workspace.workspace_dependencies import (
    is_workspace_available,
    is_workspace_owner,
)
from studio.app.common.schemas.files import FilePath, TreeNode
from studio.app.const import ACCEPT_CSV_EXT, ACCEPT_HDF5_EXT, ACCEPT_TIFF_EXT, FILETYPE
from studio.app.dir_path import DIRPATH

router = APIRouter(prefix="/files", tags=["files"])


class DirTreeGetter:
    @classmethod
    def get_tree(
        cls, workspace_id, file_types: List[str], dirname: str = None
    ) -> List[TreeNode]:
        nodes: List[TreeNode] = []

        if dirname is None:
            absolute_dirpath = join_filepath([DIRPATH.INPUT_DIR, workspace_id])
        else:
            absolute_dirpath = join_filepath([DIRPATH.INPUT_DIR, workspace_id, dirname])

        if not os.path.exists(absolute_dirpath):
            return nodes

        sorted_listdir = sorted(
            os.listdir(absolute_dirpath),
            key=lambda x: (not os.path.isdir(join_filepath([absolute_dirpath, x])), x),
        )
        for node_name in sorted_listdir:
            if dirname is None:
                relative_path = node_name
            else:
                relative_path = join_filepath([dirname, node_name])

            search_dirpath = join_filepath([absolute_dirpath, node_name])

            if os.path.isfile(search_dirpath) and node_name.endswith(tuple(file_types)):
                nodes.append(
                    TreeNode(
                        path=relative_path,
                        name=node_name,
                        isdir=False,
                        nodes=[],
                    )
                )
            elif (
                os.path.isdir(search_dirpath)
                and len(cls.accept_files(search_dirpath, file_types)) > 0
            ):
                nodes.append(
                    TreeNode(
                        path=node_name,
                        name=node_name,
                        isdir=True,
                        nodes=cls.get_tree(workspace_id, file_types, relative_path),
                    )
                )

        return nodes

    @classmethod
    def accept_files(cls, path: str, file_types: List[str]):
        files_list = []
        for file_type in file_types:
            files_list.extend(
                glob(join_filepath([path, "**", f"*{file_type}"]), recursive=True)
            )

        return files_list


@router.get(
    "/{workspace_id}",
    response_model=List[TreeNode],
    dependencies=[Depends(is_workspace_available)],
)
async def get_files(workspace_id: str, file_type: str = None):
    if file_type == FILETYPE.IMAGE:
        return DirTreeGetter.get_tree(workspace_id, ACCEPT_TIFF_EXT)
    elif file_type == FILETYPE.CSV:
        return DirTreeGetter.get_tree(workspace_id, ACCEPT_CSV_EXT)
    elif file_type == FILETYPE.HDF5:
        return DirTreeGetter.get_tree(workspace_id, ACCEPT_HDF5_EXT)
    else:
        return []


@router.post(
    "/{workspace_id}/upload/{filename}",
    response_model=FilePath,
    dependencies=[Depends(is_workspace_owner)],
)
async def create_file(workspace_id: str, filename: str, file: UploadFile = File(...)):
    create_directory(join_filepath([DIRPATH.INPUT_DIR, workspace_id]))

    filepath = join_filepath([DIRPATH.INPUT_DIR, workspace_id, filename])

    with open(filepath, "wb") as f:
        shutil.copyfileobj(file.file, f)

    return {"file_path": filename}



import os
import shutil
from dataclasses import asdict
from glob import glob
from typing import Dict

from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import FileResponse

from studio.app.common.core.experiment.experiment import ExptConfig
from studio.app.common.core.experiment.experiment_reader import ExptConfigReader
from studio.app.common.core.experiment.experiment_utils import ExptUtils
from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.workflow.workflow_reader import WorkflowConfigReader
from studio.app.common.core.workspace.workspace_dependencies import (
    is_workspace_available,
    is_workspace_owner,
)
from studio.app.common.schemas.experiment import (
    DeleteItem,
    FetchExptResponse,
    RenameItem,
)
from studio.app.dir_path import DIRPATH

router = APIRouter(prefix="/experiments", tags=["experiments"])


@router.get(
    "/{workspace_id}",
    response_model=Dict[str, ExptConfig],
    dependencies=[Depends(is_workspace_available)],
)
async def get_experiments(workspace_id: str):
    exp_config = {}
    config_paths = glob(
        join_filepath([DIRPATH.OUTPUT_DIR, workspace_id, "*", DIRPATH.EXPERIMENT_YML])
    )
    for path in config_paths:
        try:
            config = ExptConfigReader.read(path)
            exp_config[config.unique_id] = config
        except Exception:
            pass

    return exp_config


@router.patch(
    "/{workspace_id}/{unique_id}/rename",
    response_model=ExptConfig,
    dependencies=[Depends(is_workspace_owner)],
)
async def rename_experiment(workspace_id: str, unique_id: str, item: RenameItem):
    config = ExptConfigReader.rename(
        join_filepath(
            [DIRPATH.OUTPUT_DIR, workspace_id, unique_id, DIRPATH.EXPERIMENT_YML]
        ),
        new_name=item.new_name,
    )
    config.nodeDict = []
    config.edgeDict = []

    return config


@router.delete(
    "/{workspace_id}/{unique_id}",
    response_model=bool,
    dependencies=[Depends(is_workspace_owner)],
)
async def delete_experiment(workspace_id: str, unique_id: str):
    try:
        shutil.rmtree(join_filepath([DIRPATH.OUTPUT_DIR, workspace_id, unique_id]))
        return True
    except Exception:
        return False


@router.post(
    "/delete/{workspace_id}",
    response_model=bool,
    dependencies=[Depends(is_workspace_owner)],
)
async def delete_experiment_list(workspace_id: str, deleteItem: DeleteItem):
    try:
        [
            shutil.rmtree(join_filepath([DIRPATH.OUTPUT_DIR, workspace_id, uid]))
            for uid in deleteItem.uidList
        ]
        return True
    except Exception:
        return False


@router.get(
    "/fetch/{workspace_id}",
    response_model=FetchExptResponse,
    dependencies=[Depends(is_workspace_available)],
)
async def fetch_last_experiment(workspace_id: str):
    last_expt_config = ExptUtils.get_last_experiment(workspace_id)
    if last_expt_config:
        workflow_config_path = join_filepath(
            [
                DIRPATH.OUTPUT_DIR,
                workspace_id,
                last_expt_config.unique_id,
                DIRPATH.WORKFLOW_YML,
            ]
        )
        workflow_config = WorkflowConfigReader.read(workflow_config_path)
        return FetchExptResponse(**asdict(last_expt_config), **asdict(workflow_config))
    else:
        raise HTTPException(status_code=404)


@router.get(
    "/download/config/{workspace_id}/{unique_id}",
    dependencies=[Depends(is_workspace_available)],
)
async def download_config_experiment(workspace_id: str, unique_id: str):
    config_filepath = join_filepath(
        [DIRPATH.OUTPUT_DIR, workspace_id, unique_id, DIRPATH.SNAKEMAKE_CONFIG_YML]
    )
    if os.path.exists(config_filepath):
        return FileResponse(config_filepath)
    else:
        raise HTTPException(status_code=404, detail="file not found")



import pandas as pd
import plotly.express as px
import plotly.io as pio

from studio.app.common.core.utils.filepath_creater import join_filepath
from studio.app.common.core.utils.json_writer import JsonWriter
from studio.app.common.core.workflow.workflow import OutputPath, OutputType
from studio.app.common.dataclass.base import BaseData


class LineData(BaseData):
    def __init__(self, data, columns, file_name="line"):
        super().__init__(file_name)

        self.data = data
        self.columns = columns

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.json"])
        df = pd.DataFrame(self.data, columns=self.columns)
        JsonWriter.write_as_split(self.json_path, df)

    @property
    def output_path(self) -> OutputPath:
        return OutputPath(path=self.json_path, type=OutputType.LINE)

    def save_plot(self, output_dir):
        for i in range(len(self.data)):
            fig = px.line(y=self.data[i], x=self.columns)
            pio.write_image(
                fig, join_filepath([output_dir, f"{self.file_name}_{i}.png"])
            )



from typing import Dict

import yaml

from studio.app.common.core.workflow.workflow import (
    Edge,
    Node,
    NodeData,
    NodePosition,
    Style,
    WorkflowConfig,
)


class WorkflowConfigReader:
    @classmethod
    def read(cls, filepath) -> WorkflowConfig:
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)

        return WorkflowConfig(
            nodeDict=cls.read_nodeDict(config["nodeDict"]),
            edgeDict=cls.read_edgeDict(config["edgeDict"]),
        )

    @classmethod
    def read_nodeDict(cls, config) -> Dict[str, Node]:
        return {
            key: Node(
                id=key,
                type=value["type"],
                data=NodeData(**value["data"]),
                position=NodePosition(**value["position"]),
                style=Style(**value["style"]),
            )
            for key, value in config.items()
        }

    @classmethod
    def read_edgeDict(cls, config) -> Dict[str, Edge]:
        return {
            key: Edge(
                id=key,
                type=value["type"],
                animated=value["animated"],
                source=value["source"],
                sourceHandle=value["sourceHandle"],
                target=value["target"],
                targetHandle=value["targetHandle"],
                style=Style(**value["style"]),
            )
            for key, value in config.items()
        }



from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from studio.app.common.schemas.users import UserInfo


class Workspace(BaseModel):
    id: Optional[int]
    name: str
    user: Optional[UserInfo]
    shared_count: int
    created_at: Optional[datetime]
    updated_at: Optional[datetime]

    class Config:
        orm_mode = True


class WorkspaceCreate(BaseModel):
    name: str


class WorkspaceUpdate(WorkspaceCreate):
    pass


class WorkspaceSharePostStatus(BaseModel):
    user_ids: Optional[List[int]]


class WorkspaceShareStatus(BaseModel):
    users: Optional[List[UserInfo]]



from studio.app.common.models.user import Organization, Role, User, UserRole
from studio.app.common.models.workspace import Workspace, WorkspacesShareUser

__all__ = [
    "Organization",
    "Role",
    "User",
    "UserRole",
    "Workspace",
    "WorkspacesShareUser",
]



from sqlalchemy.orm import sessionmaker
from sqlmodel import create_engine

from studio.app.common.db.config import DATABASE_CONFIG

engine = create_engine(DATABASE_CONFIG.DATABASE_URL, pool_recycle=360, pool_size=20)
SessionLocal = sessionmaker(
    autocommit=False, autoflush=False, expire_on_commit=False, bind=engine
)


def get_db():
    try:
        db = SessionLocal()
        yield db
    finally:
        db.close()



from typing import Dict

import yaml

from studio.app.common.core.experiment.experiment import ExptConfig, ExptFunction
from studio.app.common.core.workflow.workflow import OutputPath


class ExptConfigReader:
    @classmethod
    def read(cls, filepath) -> ExptConfig:
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)

        return ExptConfig(
            workspace_id=config["workspace_id"],
            unique_id=config["unique_id"],
            name=config["name"],
            started_at=config["started_at"],
            finished_at=config.get("finished_at"),
            success=config.get("success", "running"),
            hasNWB=config["hasNWB"],
            function=cls.read_function(config["function"]),
            nwb=config.get("nwb"),
            snakemake=config.get("snakemake"),
        )

    @classmethod
    def read_function(cls, config) -> Dict[str, ExptFunction]:
        return {
            key: ExptFunction(
                unique_id=value["unique_id"],
                name=value["name"],
                started_at=value.get("started_at"),
                finished_at=value.get("finished_at"),
                success=value.get("success", "running"),
                hasNWB=value["hasNWB"],
                message=value.get("message"),
                outputPaths=cls.read_output_paths(value.get("outputPaths")),
            )
            for key, value in config.items()
        }

    @classmethod
    def read_output_paths(cls, config) -> Dict[str, OutputPath]:
        if config:
            return {
                key: OutputPath(
                    path=value["path"],
                    type=value["type"],
                    max_index=value["max_index"],
                )
                for key, value in config.items()
            }
        else:
            return None

    @classmethod
    def rename(cls, filepath, new_name: str) -> ExptConfig:
        with open(filepath, "r") as f:
            config = yaml.safe_load(f)
            config["name"] = new_name

        with open(filepath, "w") as f:
            yaml.dump(config, f)

        return ExptConfig(
            workspace_id=config["workspace_id"],
            unique_id=config["unique_id"],
            name=config["name"],
            started_at=config.get("started_at"),
            finished_at=config.get("finished_at"),
            success=config.get("success", "running"),
            hasNWB=config["hasNWB"],
            function=cls.read_function(config["function"]),
            nwb=config.get("nwb"),
            snakemake=config.get("snakemake"),
        )



import gc

import imageio
import numpy as np
import tifffile

from studio.app.common.core.utils.filepath_creater import (
    create_directory,
    join_filepath,
)
from studio.app.common.core.utils.json_writer import JsonWriter
from studio.app.common.core.workflow.workflow import OutputPath, OutputType
from studio.app.common.dataclass.base import BaseData
from studio.app.common.dataclass.utils import create_images_list
from studio.app.dir_path import DIRPATH


class RoiData(BaseData):
    def __init__(self, data, output_dir=DIRPATH.OUTPUT_DIR, file_name="roi"):
        super().__init__(file_name)

        images = create_images_list(data)

        _dir = join_filepath([output_dir, "tiff", file_name])
        create_directory(_dir)
        self.path = join_filepath([_dir, f"{file_name}.tif"])
        tifffile.imsave(self.path, images)

        del images, data
        gc.collect()

    @property
    def data(self):
        data = np.array(imageio.volread(self.path))
        if data.ndim == 3:
            return data[0]
        elif data.ndim == 2:
            return data

    def save_json(self, json_dir):
        self.json_path = join_filepath([json_dir, f"{self.file_name}.json"])
        JsonWriter.write_as_split(self.json_path, create_images_list(self.data))

    @property
    def output_path(self) -> OutputPath:
        return OutputPath(path=self.json_path, type=OutputType.ROI)



from studio.app.optinist.wrappers.optinist.neural_decoding.glm import GLM
from studio.app.optinist.wrappers.optinist.neural_decoding.lda import LDA
from studio.app.optinist.wrappers.optinist.neural_decoding.svm import SVM

neural_decoding_wrapper_dict = {
    "glm": {
        "function": GLM,
        "conda_name": "optinist",
    },
    "lda": {
        "function": LDA,
        "conda_name": "optinist",
    },
    "svm": {
        "function": SVM,
        "conda_name": "optinist",
    },
}



import numpy as np
import pandas as pd

from studio.app.common.dataclass import HeatMapData  # , TimeSeriesData
from studio.app.optinist.core.nwb.nwb import NWBDATASET
from studio.app.optinist.dataclass import BehaviorData, FluoData, IscellData
from studio.app.optinist.wrappers.optinist.utils import standard_norm


def calc_trigger(behavior_data, trigger_type, trigger_threshold):
    # same function is also in the eta
    flg = np.array(behavior_data > trigger_threshold, dtype=int)
    if trigger_type == "up":
        trigger_idx = np.where(np.ediff1d(flg) == 1)[0]
    elif trigger_type == "down":
        trigger_idx = np.where(np.ediff1d(flg) == -1)[0]
    elif trigger_type == "cross":
        trigger_idx = np.where(np.ediff1d(flg) != 0)[0]
    else:
        trigger_idx = np.where(np.ediff1d(flg) == 0)[0]

    return trigger_idx


def GetIndices(dims, outtype):
    index = np.indices(dims)
    ind = []
    for i in range(len(index)):
        ind.append(index[i].flatten())
    ind = np.array(ind)
    ind = ind.transpose()

    if outtype == "list":
        out = []
        for i in range(ind.shape[1]):
            out.append(ind[:, i])
    else:
        out = ind

    return out


def createMatrix(D, triggers, stims, duration):
    # D: num_timestamps x num_unit   neural data
    # triggers: index of trigger
    # stims: list of stimulus property for each trigger
    # durations: frames before and after trigger to use

    num_unit = D.shape[1]
    num_triggers = len(triggers)
    num_property = len(stims)
    num_timepoints = duration[1] - duration[0]

    # X is the reshaped data for each trigger
    X = np.zeros([num_triggers, num_unit, num_timepoints])
    for i in range(num_triggers):
        X[i, :, :] = D[
            triggers[i] + duration[0] : triggers[i] + duration[1], :
        ].transpose()

    # df is the table of stimulus conditions
    # df: number of trigger x ( number of property +1 )
    # uq_stims = list of unique stimulus for each property
    # num_uq_stims = number of unique_stims for each property

    columns = columns = list(map(str, range(num_property)))
    columns.append("all")
    df = pd.DataFrame(data=None, index=list(range(num_triggers)), columns=columns)
    for i in range(num_property):
        df[columns[i]] = stims[i]

    for i in range(num_triggers):
        df.iloc[i, num_property] = "_".join(map(str, df.iloc[i, 0:2].values.tolist()))

    uq_list = df["all"].unique()
    uq_stims = []
    num_uq_stims = []
    for i in range(num_property):
        uq_stims.append(list(df.iloc[:, i].unique()))
        num_uq_stims.append(len(uq_stims[i]))

    #  check number of samples
    # n: number of samples for each condition
    # index: index of the trigger for each condition
    # min_sample: minimum number of samples for a condition
    n = np.zeros([len(uq_list)], dtype=int)
    index = []
    for i in range(len(uq_list)):
        index.append(list(df[df["all"] == uq_list[i]].index))
        n[i] = len(index[i])

    min_sample = int(np.min(n))

    # re-format the data (number of samples is set to the nun_sample)
    X2 = np.zeros([min_sample, num_unit, num_timepoints] + num_uq_stims)

    for i in range(len(index)):
        stims = list(df.iloc[index[i][0], 0 : df.shape[1] - 1])
        tgtind = []
        for j in range(len(stims)):
            tgtind.append(uq_stims[j].index(stims[j]))

        if num_property == 1:
            X2[0:min_sample, :, :, tgtind[0]] = X[index[i][0:min_sample], :, :]
        elif num_property == 2:
            X2[0:min_sample, :, :, tgtind[0], tgtind[1]] = X[
                index[i][0:min_sample], :, :
            ]
        elif num_property == 3:
            X2[0:min_sample, :, :, tgtind[0], tgtind[1], tgtind[2]] = X[
                index[i][0:min_sample], :, :
            ]

        else:
            print("currently the number of condition category has to be less than 4")
            return

    return X2


def reshapeBehavior(
    B, trigger_column, trigger_type, trigger_threshold, feature_columns
):
    Trig = calc_trigger(B[:, trigger_column], trigger_type, trigger_threshold)

    features = []
    for i in range(len(feature_columns)):
        features.append(B[Trig, feature_columns[i]])

    return [Trig, features]


def dpca_fit(
    neural_data: FluoData,
    behaviors_data: BehaviorData,
    output_dir: str,
    iscell: IscellData = None,
    params: dict = None,
    **kwargs,
) -> dict():
    # modules specific to function
    from dPCA import dPCA

    function_id = output_dir.split("/")[-1]
    print("start dpca:", function_id)

    neural_data = neural_data.data
    behaviors_data = behaviors_data.data

    # neural data shold be time x cells
    if params["transpose"]:
        X = neural_data.transpose()
    else:
        X = neural_data

    if iscell is not None:
        iscell = iscell.data
        ind = np.where(iscell > 0)[0]
        X = X[:, ind]

    # preprocessing
    X = standard_norm(X, params["standard_mean"], params["standard_std"])

    # create trigger and features
    [Trig, features] = reshapeBehavior(
        behaviors_data,
        params["trigger_column"],
        params["trigger_type"],
        params["trigger_threshold"],
        params["feature_colums"],
    )
    X = createMatrix(X, Trig, features, params["trigger_duration"])

    # calculate dPCA  #

    # X: array - like, shape(n_samples, n_features_1, n_features_2, ...)
    # Training data, where n_samples in the number of samples
    # and n_features_j is the number
    # of the j - features(where the axis correspond to different parameters).

    dpca = dPCA.dPCA(
        labels=params["labels"],
        join=params["join"],
        regularizer=params["regularizer"],
        n_components=params["n_components"],
        copy=params["copy"],
        n_iter=params["n_iter"],
    )

    result = dpca.fit_transform(np.mean(X, axis=0), X)
    keys = list(result.keys())

    Out_forfigure = {}
    # figure shows only assigned components and properties
    for i in range(len(params["figure_features"])):
        for j in range(len(params["figure_components"])):
            tp = result[params["figure_features"][i]][
                params["figure_components"][j],
            ]  # 1st component
            inds = GetIndices(tp.shape[1:], "matrix")
            arr = np.zeros([tp.shape[0], inds.shape[0]])
            for m in range(inds.shape[0]):
                for k in range(tp.shape[0]):
                    arr[k, m] = tp[tuple([k] + list(inds[m, :]))]

            Out_forfigure[
                params["figure_features"][i]
                + "-component"
                + str(params["figure_components"][j])
            ] = arr
    Out_forfigure["features"] = list(Out_forfigure.keys())

    names = []
    for i in range(inds.shape[0]):
        names.append("feature" + "_".join(map(str, inds[i, :])))
    Out_forfigure["trace_names"] = names

    # NWB
    tpdic = {}
    for i in range(len(keys)):
        tpdic[keys[i]] = result[keys[i]]

    nwbfile = {}
    nwbfile[NWBDATASET.POSTPROCESS] = {function_id: {**tpdic}}

    info = {}
    for i in range(len(Out_forfigure["features"])):
        # info[Out_forfigure["features"][i]] = TimeSeriesData(
        #     Out_forfigure[Out_forfigure["features"][i]].transpose(),
        #     std=None,
        #     index=None,
        #     file_name=Out_forfigure["features"][i],
        # )

        info[Out_forfigure["features"][i]] = HeatMapData(
            Out_forfigure[Out_forfigure["features"][i]].transpose(),
            columns=None,
            file_name=Out_forfigure["features"][i],
        )
    info["nwbfile"] = nwbfile

    return info



from typing import Any, Dict, Optional

from pydantic import BaseSettings, Field, validator

from studio.app.dir_path import DIRPATH


class DatabaseConfig(BaseSettings):
    """configuration for db"""

    MYSQL_ROOT_PASSWORD: str = Field(default=True, env="MYSQL_ROOT_PASSWORD")
    MYSQL_SERVER: str = Field(default="db", env="MYSQL_SERVER")
    MYSQL_USER: str = Field(default=None, env="MYSQL_USER")
    MYSQL_PASSWORD: str = Field(default=None, env="MYSQL_PASSWORD")
    MYSQL_DATABASE: str = Field(default=None, env="MYSQL_DATABASE")
    DATABASE_URL: str = Field(default=None)

    @validator("DATABASE_URL", pre=True)
    def assemble_db_connection(cls, v: Optional[str], values: Dict[str, Any]) -> Any:
        if isinstance(v, str):
            return v
        user = values.get("MYSQL_USER")
        password = values.get("MYSQL_PASSWORD")
        server = values.get("MYSQL_SERVER")
        database = values.get("MYSQL_DATABASE")

        return f"mysql+pymysql://{user}:{password}@{server}/{database}?charset=utf8mb4"

    class Config:
        env_file = f"{DIRPATH.CONFIG_DIR}/.env"
        env_file_encoding = "utf-8"
        case_sensitive = True


DATABASE_CONFIG = DatabaseConfig()
